import datetime
import os
import time
import threading

from typing import Optional, Union, Dict, Any, List

import croniter
import enum

from platypush.backend import Backend
from platypush.context import get_bus, get_plugin
from platypush.message.event.alarm import AlarmStartedEvent, AlarmDismissedEvent, AlarmSnoozedEvent
from platypush.plugins.media import MediaPlugin, PlayerState
from platypush.procedure import Procedure


class AlarmState(enum.IntEnum):
    WAITING = 1
    RUNNING = 2
    DISMISSED = 3
    SNOOZED = 4
    SHUTDOWN = 5


class Alarm:
    _alarms_count = 0
    _id_lock = threading.RLock()

    def __init__(self, when: str, actions: Optional[list] = None, name: Optional[str] = None,
                 audio_file: Optional[str] = None, audio_plugin: Optional[str] = None,
                 audio_volume: Optional[Union[int, float]] = None,
                 snooze_interval: float = 300.0, enabled: bool = True):
        with self._id_lock:
            self._alarms_count += 1
            self.id = self._alarms_count

        self.when = when
        self.name = name or 'Alarm_{}'.format(self.id)
        self.audio_file = None

        if audio_file:
            self.audio_file = os.path.abspath(os.path.expanduser(audio_file))
            assert os.path.isfile(self.audio_file), 'No such audio file: {}'.format(self.audio_file)

        self.audio_plugin = audio_plugin
        self.audio_volume = audio_volume
        self.snooze_interval = snooze_interval
        self.state: Optional[AlarmState] = None
        self.timer: Optional[threading.Timer] = None
        self.actions = Procedure.build(name=name, _async=False, requests=actions or [], id=self.id)

        self._enabled = enabled
        self._runtime_snooze_interval = snooze_interval

    def get_next(self) -> float:
        now = time.time()

        try:
            cron = croniter.croniter(self.when, now)
            return cron.get_next()
        except (AttributeError, croniter.CroniterBadCronError):
            try:
                timestamp = datetime.datetime.fromisoformat(self.when).timestamp()
            except (TypeError, ValueError):
                timestamp = (datetime.datetime.now() + datetime.timedelta(seconds=int(self.when))).timestamp()

            return timestamp if timestamp >= now else None

    def is_enabled(self):
        return self._enabled

    def disable(self):
        self._enabled = False

    def enable(self):
        self._enabled = True

    def dismiss(self):
        self.state = AlarmState.DISMISSED
        self.stop_audio()
        get_bus().post(AlarmDismissedEvent(name=self.name))

    def snooze(self, interval: Optional[float] = None):
        self._runtime_snooze_interval = interval or self.snooze_interval
        self.state = AlarmState.SNOOZED
        self.stop_audio()
        get_bus().post(AlarmSnoozedEvent(name=self.name, interval=self._runtime_snooze_interval))

    def start(self):
        if self.timer:
            self.timer.cancel()

        if self.get_next() is None:
            return

        interval = self.get_next() - time.time()
        self.timer = threading.Timer(interval, self.callback())
        self.timer.start()
        self.state = AlarmState.WAITING

    def stop(self):
        self.state = AlarmState.SHUTDOWN
        if self.timer:
            self.timer.cancel()
            self.timer = None

    def _get_audio_plugin(self) -> MediaPlugin:
        return get_plugin(self.audio_plugin)

    def play_audio(self):
        def thread():
            self._get_audio_plugin().play(self.audio_file)
            if self.audio_volume is not None:
                self._get_audio_plugin().set_volume(self.audio_volume)

        self.state = AlarmState.RUNNING
        audio_thread = threading.Thread(target=thread)
        audio_thread.start()

    def stop_audio(self):
        self._get_audio_plugin().stop()

    def callback(self):
        def _callback():
            while True:
                if self.state == AlarmState.SHUTDOWN:
                    break

                if self.is_enabled():
                    get_bus().post(AlarmStartedEvent(name=self.name))
                    if self.audio_plugin and self.audio_file:
                        self.play_audio()

                    self.actions.execute()

                time.sleep(10)
                sleep_time = None
                if self.state == AlarmState.RUNNING:
                    while True:
                        state = self._get_audio_plugin().status().output.get('state')
                        if state == PlayerState.STOP.value:
                            if self.state == AlarmState.SNOOZED:
                                sleep_time = self._runtime_snooze_interval
                            else:
                                self.state = AlarmState.WAITING

                            break
                        else:
                            time.sleep(10)

                if self.state == AlarmState.SNOOZED:
                    sleep_time = self._runtime_snooze_interval
                elif self.get_next() is None:
                    self.state = AlarmState.SHUTDOWN
                    break

                if not sleep_time:
                    sleep_time = self.get_next() - time.time() if self.get_next() else 10

                time.sleep(sleep_time)

        return _callback

    def to_dict(self):
        return {
            'name': self.name,
            'id': self.id,
            'when': self.when,
            'next_run': self.get_next(),
            'enabled': self.is_enabled(),
            'state': self.state.name,
        }


class AlarmBackend(Backend):
    """
    Backend to handle user-configured alarms.

    Triggers:

        * :class:`platypush.message.event.alarm.AlarmStartedEvent` when an alarm starts.
        * :class:`platypush.message.event.alarm.AlarmSnoozedEvent` when an alarm is snoozed.
        * :class:`platypush.message.event.alarm.AlarmTimeoutEvent` when an alarm times out.
        * :class:`platypush.message.event.alarm.AlarmDismissedEvent` when an alarm is dismissed.

    """

    def __init__(self, alarms: Optional[Union[list, Dict[str, Any]]] = None, audio_plugin: str = 'media.mplayer',
                 *args, **kwargs):
        """
        :param alarms: List or name->value dict with the configured alarms. Example:

        .. code-block:: yaml

            morning_alarm:
                when: '0 7 * * 1-5'   # Cron expression format: run every weekday at 7 AM
                audio_file: ~/path/your_ringtone.mp3
                audio_plugin: media.mplayer
                audio_volume: 10       # 10%
                snooze_interval: 300   # 5 minutes snooze
                actions:
                    - action: tts.say
                      args:
                          text: Good morning

                    - action: light.hue.bri
                      args:
                          value: 1

                    - action: light.hue.bri
                      args:
                          value: 140
                          transitiontime: 150

            one_shot_alarm:
                when: '2020-02-18T07:00:00.000000'   # One-shot execution, with timestamp in ISO format
                audio_file: ~/path/your_ringtone.mp3
                actions:
                    - action: light.hue.on

        :param audio_plugin: Media plugin (instance of :class:`platypush.plugins.media.MediaPlugin`) that will be
            used to play the alarm audio (default: ``media.mplayer``).
        """
        super().__init__(*args, **kwargs)
        alarms = alarms or []
        if isinstance(alarms, dict):
            alarms = [{'name': name, **alarm} for name, alarm in alarms.items()]

        self.audio_plugin = audio_plugin
        alarms = [Alarm(**{'audio_plugin': self.audio_plugin, **alarm}) for alarm in alarms]
        self.alarms: Dict[str, Alarm] = {alarm.name: alarm for alarm in alarms}

    def add_alarm(self, when: str, actions: list, name: Optional[str] = None, audio_file: Optional[str] = None,
                  audio_volume: Optional[Union[int, float]] = None, enabled: bool = True) -> Alarm:
        alarm = Alarm(when=when, actions=actions, name=name, enabled=enabled, audio_file=audio_file,
                      audio_plugin=self.audio_plugin, audio_volume=audio_volume)

        if alarm.name in self.alarms:
            self.logger.info('Overwriting existing alarm {}'.format(alarm.name))
            self.alarms[alarm.name].stop()

        self.alarms[alarm.name] = alarm
        self.alarms[alarm.name].start()
        return self.alarms[alarm.name]

    def _get_alarm(self, name) -> Alarm:
        assert name in self.alarms, 'Alarm {} does not exist'.format(name)
        return self.alarms[name]

    def enable_alarm(self, name: str):
        self._get_alarm(name).enable()

    def disable_alarm(self, name: str):
        self._get_alarm(name).disable()

    def dismiss_alarm(self):
        alarm = self.get_running_alarm()
        if not alarm:
            self.logger.info('No alarm is running')
            return

        alarm.dismiss()

    def snooze_alarm(self, interval: Optional[str] = None):
        alarm = self.get_running_alarm()
        if not alarm:
            self.logger.info('No alarm is running')
            return

        alarm.snooze(interval=interval)

    def get_alarms(self) -> List[Alarm]:
        return sorted([alarm for alarm in self.alarms.values()], key=lambda alarm: alarm.get_next())

    def get_running_alarm(self) -> Optional[Alarm]:
        running_alarms = [alarm for alarm in self.alarms.values() if alarm.state == AlarmState.RUNNING]
        return running_alarms[0] if running_alarms else None

    def __enter__(self):
        for alarm in self.alarms.values():
            alarm.stop()
            alarm.start()

        self.logger.info('Initialized alarm backend with {} alarms'.format(len(self.alarms)))

    def __exit__(self, exc_type, exc_val, exc_tb):
        for alarm in self.alarms.values():
            alarm.stop()

        self.logger.info('Alarm backend terminated')

    def loop(self):
        for name, alarm in self.alarms.copy().items():
            if not alarm.timer or (not alarm.timer.is_alive() and alarm.state == AlarmState.SHUTDOWN):
                del self.alarms[name]

        time.sleep(10)


# vim:sw=4:ts=4:et:



import base64
import os
import subprocess
import sys
import time
from typing import Optional, Dict

from platypush.plugins import action
from platypush.plugins.sensor import SensorPlugin
from platypush.message.response.bluetooth import BluetoothScanResponse, BluetoothDiscoverPrimaryResponse, \
    BluetoothDiscoverCharacteristicsResponse


class BluetoothBlePlugin(SensorPlugin):
    """
    Bluetooth BLE (low-energy) plugin

    Requires:

        * **pybluez** (``pip install pybluez``)
        * **gattlib** (``pip install gattlib``)

    Note that the support for bluetooth low-energy devices on Linux requires:

        * A bluetooth adapter compatible with the bluetooth 5.0 specification or higher;
        * To run platypush with root privileges (which is usually a very bad idea), or to set the raw net
          capabilities on the Python executable (which is also a bad idea, because any Python script will
          be able to access the kernel raw network API, but it's probably better than running a network
          server that can execute system commands as root user). If you don't want to set special permissions
          on the main Python executable and you want to run the bluetooth LTE plugin then the advised approach
          is to install platypush in a virtual environment and set the capabilities on the venv python executable,
          or run your platypush instance in Docker.

          You can set the capabilities on the Python executable through the following shell command::

            [sudo] setcap 'cap_net_raw,cap_net_admin+eip' /path/to/your/python

    """

    def __init__(self, interface: str = 'hci0', **kwargs):
        """
        :param interface: Default adapter device to be used (default: 'hci0')
        """
        super().__init__(**kwargs)
        self.interface = interface
        self._req_by_addr = {}

    @staticmethod
    def _get_python_interpreter() -> str:
        exe = sys.executable

        while os.path.islink(exe):
            target = os.readlink(exe)
            if not os.path.isabs(target):
                target = os.path.abspath(os.path.join(os.path.dirname(exe), target))
            exe = target

        return exe

    @staticmethod
    def _python_has_ble_capabilities(exe: str) -> bool:
        getcap = subprocess.Popen(['getcap', exe], stdout=subprocess.PIPE)
        output = getcap.communicate()[0].decode().split('\n')
        if not output:
            return False

        caps = set(output.pop(0).split('=').pop().strip().split(','))
        return 'cap_net_raw+eip' in caps and 'cap_net_admin' in caps

    def _check_ble_support(self):
        # Check if the script is running as root or if the Python executable
        # has 'cap_net_admin,cap_net_raw+eip' capabilities
        exe = self._get_python_interpreter()
        if os.getuid() != 0 and not self._python_has_ble_capabilities(exe):
            raise RuntimeError('You are not running platypush as root and the Python interpreter has no ' +
                               'capabilities/permissions to access the BLE stack. Set the permissions on ' +
                               'your Python interpreter through:\n' +
                               '\t[sudo] setcap "cap_net_raw,cap_net_admin+eip" {}'.format(exe))

    @action
    def scan(self, interface: Optional[str] = None, duration: int = 10) -> BluetoothScanResponse:
        """
        Scan for nearby bluetooth low-energy devices

        :param interface: Bluetooth adapter name to use (default configured if None)
        :param duration: Scan duration in seconds
        """
        from bluetooth.ble import DiscoveryService

        if interface is None:
            interface = self.interface

        self._check_ble_support()
        svc = DiscoveryService(interface)
        devices = svc.discover(duration)
        return BluetoothScanResponse(devices)

    @action
    def get_measurement(self, interface: Optional[str] = None, duration: Optional[int] = 10, *args, **kwargs) \
            -> Dict[str, dict]:
        """
        Wrapper for ``scan`` that returns bluetooth devices in a format usable by sensor backends.

        :param interface: Bluetooth adapter name to use (default configured if None)
        :param duration: Scan duration in seconds
        :return: Device address -> info map.
        """
        devices = self.scan(interface=interface, duration=duration).output
        return {device['addr']: device for device in devices}

    # noinspection PyArgumentList
    @action
    def connect(self, device: str, interface: str = None, wait: bool = True, channel_type: str = 'public',
                security_level: str = 'low', psm: int = 0, mtu: int = 0, timeout: float = 10.0):
        """
        Connect to a bluetooth LE device

        :param device: Device address to connect to
        :param interface: Bluetooth adapter name to use (default configured if None)
        :param wait: If True then wait for the connection to be established before returning (no timeout)
        :param channel_type: Channel type, usually 'public' or 'random'
        :param security_level: Security level - possible values: ['low', 'medium', 'high']
        :param psm: PSM value (default: 0)
        :param mtu: MTU value (default: 0)
        :param timeout: Connection timeout if wait is not set (default: 10 seconds)
        """
        from gattlib import GATTRequester

        req = self._req_by_addr.get(device)
        if req:
            if req.is_connected():
                self.logger.info('Device {} is already connected'.format(device))
                return

            self._req_by_addr[device] = None

        if not interface:
            interface = self.interface
        if interface:
            req = GATTRequester(device, False, interface)
        else:
            req = GATTRequester(device, False)

        self.logger.info('Connecting to {}'.format(device))
        connect_start_time = time.time()
        req.connect(wait, channel_type, security_level, psm, mtu)

        if not wait:
            while not req.is_connected():
                if time.time() - connect_start_time > timeout:
                    raise TimeoutError('Connection to {} timed out'.format(device))
                time.sleep(0.1)

        self.logger.info('Connected to {}'.format(device))
        self._req_by_addr[device] = req

    @action
    def read(self, device: str, interface: str = None, uuid: str = None, handle: int = None,
             binary: bool = False, disconnect_on_recv: bool = True, **kwargs) -> str:
        """
        Read a message from a device

        :param device: Device address to connect to
        :param interface: Bluetooth adapter name to use (default configured if None)
        :param uuid: Service UUID. Either the UUID or the device handle must be specified
        :param handle: Device handle. Either the UUID or the device handle must be specified
        :param binary: Set to true to return data as a base64-encoded binary string
        :param disconnect_on_recv: If True (default) disconnect when the response is received
        :param kwargs: Extra arguments to be passed to :meth:`connect`
        """
        if interface is None:
            interface = self.interface
        if not (uuid or handle):
            raise AttributeError('Specify either uuid or handle')

        self.connect(device, interface=interface, **kwargs)
        req = self._req_by_addr[device]

        if uuid:
            data = req.read_by_uuid(uuid)[0]
        else:
            data = req.read_by_handle(handle)[0]

        if binary:
            data = base64.encodebytes(data.encode() if isinstance(data, str) else data).decode().strip()
        if disconnect_on_recv:
            self.disconnect(device)

        return data

    @action
    def write(self, device: str, data, handle: int = None, interface: str = None, binary: bool = False,
              disconnect_on_recv: bool = True, **kwargs) -> str:
        """
        Writes data to a device

        :param device: Device address to connect to
        :param data: Data to be written (str or bytes)
        :param interface: Bluetooth adapter name to use (default configured if None)
        :param handle: Device handle. Either the UUID or the device handle must be specified
        :param binary: Set to true if data is a base64-encoded binary string
        :param disconnect_on_recv: If True (default) disconnect when the response is received
        :param kwargs: Extra arguments to be passed to :meth:`connect`
        """
        if interface is None:
            interface = self.interface
        if binary:
            data = base64.decodebytes(data.encode() if isinstance(data, str) else data)

        self.connect(device, interface=interface, **kwargs)
        req = self._req_by_addr[device]

        data = req.write_by_handle(handle, data)[0]

        if binary:
            data = base64.encodebytes(data.encode() if isinstance(data, str) else data).decode().strip()
        if disconnect_on_recv:
            self.disconnect(device)

        return data

    @action
    def disconnect(self, device: str):
        """
        Disconnect from a connected device

        :param device: Device address
        """
        req = self._req_by_addr.get(device)
        if not req:
            self.logger.info('Device {} not connected'.format(device))

        req.disconnect()
        self.logger.info('Device {} disconnected'.format(device))

    @action
    def discover_primary(self, device: str, interface: str = None, **kwargs) -> BluetoothDiscoverPrimaryResponse:
        """
        Discover the primary services advertised by a LE bluetooth device

        :param device: Device address to connect to
        :param interface: Bluetooth adapter name to use (default configured if None)
        :param kwargs: Extra arguments to be passed to :meth:`connect`
        """
        if interface is None:
            interface = self.interface

        self.connect(device, interface=interface, **kwargs)
        req = self._req_by_addr[device]
        services = req.discover_primary()
        self.disconnect(device)
        return BluetoothDiscoverPrimaryResponse(services=services)

    @action
    def discover_characteristics(self, device: str, interface: str = None, **kwargs) \
            -> BluetoothDiscoverCharacteristicsResponse:
        """
        Discover the characteristics of a LE bluetooth device

        :param device: Device address to connect to
        :param interface: Bluetooth adapter name to use (default configured if None)
        :param kwargs: Extra arguments to be passed to :meth:`connect`
        """
        if interface is None:
            interface = self.interface

        self.connect(device, interface=interface, **kwargs)
        req = self._req_by_addr[device]
        characteristics = req.discover_characteristics()
        self.disconnect(device)
        return BluetoothDiscoverCharacteristicsResponse(characteristics=characteristics)


# vim:sw=4:ts=4:et:



import csv
import os
from typing import Optional, List, Any, Union, Dict
from typing.io import TextIO

from platypush.plugins import Plugin, action


class CsvPlugin(Plugin):
    """
    A plugin for managing CSV files.
    """

    @classmethod
    def _get_path(cls, filename: str) -> str:
        return os.path.abspath(os.path.expanduser(filename))

    @staticmethod
    def reversed_blocks(f: TextIO, blocksize=4096):
        """ Generate blocks of file's contents in reverse order. """
        f.seek(0, os.SEEK_END)
        here = f.tell()
        while 0 < here:
            delta = min(blocksize, here)
            here -= delta
            f.seek(here, os.SEEK_SET)
            yield f.read(delta)

    @classmethod
    def lines(cls, f: TextIO, reverse: bool = False):
        if not reverse:
            for line in f:
                yield line
        else:
            part = ''
            quoting = False
            for block in cls.reversed_blocks(f):
                for c in reversed(block):
                    if c == '"':
                        quoting = not quoting
                    elif c == '\n' and part and not quoting:
                        yield part[::-1]
                        part = ''
                    part += c
            if part:
                yield part[::-1]

    @staticmethod
    def _parse_header(filename: str, **csv_args) -> List[str]:
        column_names = []
        with open(filename, 'r', newline='') as f:
            try:
                has_header = csv.Sniffer().has_header(f.read(1024))
            except csv.Error:
                has_header = False

        if has_header:
            with open(filename, 'r', newline='') as f:
                for row in csv.reader(f, **csv_args):
                    column_names = row
                    break

        return column_names

    @action
    def read(self,
             filename: str,
             delimiter: str = ',',
             quotechar: Optional[str] = '"',
             start: int = 0,
             limit: Optional[int] = None,
             reverse: bool = False,
             has_header: bool = None,
             column_names: Optional[List[str]] = None,
             dialect: str = 'excel'):
        """
        Gets the content of a CSV file.

        :param filename: Path of the file.
        :param delimiter: Field delimiter (default: ``,``).
        :param quotechar: Quote character (default: ``"``).
        :param start: (Zero-based) index of the first line to be read (starting from the last if ``reverse`` is True)
            (default: 0).
        :param limit: Maximum number of lines to be read (default: all).
        :param reverse: If True then the lines will be read starting from the last (default: False).
        :param has_header: Set to True if the first row of the file is a header, False if the first row
            isn't expected to be a header (default: None, the method will scan the first chunk of the file
            and estimate whether the first line is a header).
        :param column_names: Specify if the file has no header or you want to override the column names.
        :param dialect: CSV dialect (default: ``excel``).
        """

        filename = self._get_path(filename)
        column_names = column_names or []
        csv_args = {
            'delimiter': delimiter,
            'quotechar': quotechar,
            'dialect': dialect,
        }

        if has_header is None and not column_names:
            column_names = self._parse_header(filename, **csv_args)
            has_header = len(column_names) > 0

        rows = []
        with open(filename, 'r', newline='') as f:
            for i, row in enumerate(csv.reader(self.lines(f, reverse=reverse), **csv_args)):
                if not row or i < start:
                    continue
                if limit and len(rows) >= limit + (1 if has_header else 0):
                    break

                rows.append(dict(zip(column_names, row)) if column_names else row)

        if has_header:
            rows.pop(-1 if reverse else 0)
        return rows

    @action
    def write(self,
              filename: str,
              rows: List[Union[List[Any], Dict[str, Any]]],
              truncate: bool = False,
              delimiter: str = ',',
              quotechar: Optional[str] = '"',
              dialect: str = 'excel'):
        """
        Writes lines to a CSV file.

        :param filename: Path of the CSV file.
        :param rows: Rows to write. It can be a list of lists or a key->value dictionary where the keys match
            the names of the columns. If the rows are dictionaries then a header with the column names will be
            written to the file if not available already, otherwise no header will be written.
        :param truncate: If True then any previous file content will be removed, otherwise the new rows will be
            appended to the file (default: False).
        :param delimiter: Field delimiter (default: ``,``).
        :param quotechar: Quote character (default: ``"``).
        :param dialect: CSV dialect (default: ``excel``).
        """
        filename = self._get_path(filename)
        file_exists = os.path.isfile(filename)
        column_names = []
        csv_args = {
            'delimiter': delimiter,
            'quotechar': quotechar,
            'dialect': dialect,
        }

        if file_exists:
            column_names = self._parse_header(filename, **csv_args)
        elif rows and isinstance(rows[0], dict):
            column_names = rows[0].keys()

        column_name_to_idx = {name: i for i, name in enumerate(column_names)}
        if truncate:
            file_exists = False

        with open(filename, 'w' if truncate else 'a', newline='') as f:
            writer = csv.writer(f, **csv_args)
            if not file_exists and column_names:
                writer.writerow(column_names)

            for row in rows:
                if isinstance(row, dict):
                    flat_row = [None] * len(column_names)
                    for column, value in row.items():
                        assert column in column_name_to_idx, \
                            'No such column available in the CSV file: {}'.format(column)
                        idx = column_name_to_idx[column]
                        flat_row[idx] = value

                    row = flat_row

                writer.writerow(row)


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class MQTTMessageEvent(Event):
    """
    MQTT message event object. Fired when :mod:`platypush.backend.mqtt` receives
    a new event.
    """

    def __init__(self, msg, host=None, port=None, topic=None, *args, **kwargs):
        super().__init__(msg=msg, host=host, port=port, topic=topic,
                         *args, **kwargs)


# vim:sw=4:ts=4:et:



import os

from platypush.plugins import Plugin, action


class FilePlugin(Plugin):
    """
    A plugin for general-purpose file methods
    """

    @classmethod
    def _get_path(cls, filename):
        return os.path.abspath(os.path.expanduser(filename))

    @action
    def get(self, filename):
        """
        Gets the content of a file

        :param filename: Path of the file
        :type filename: str
        """

        with open(self._get_path(filename), 'r') as f:
            return f.read()

    @action
    def write(self, filename, content):
        """
        Writes content to a specified filename. Previous content will be truncated.

        :param filename: Path of the file
        :type filename: str

        :param content: Content to write
        :type content: str
        """

        with open(self._get_path(filename), 'w') as f:
            f.write(content)

    @action
    def append(self, filename, content):
        """
        Append content to a specified filename

        :param filename: Path of the file
        :type filename: str

        :param content: Content to write
        :type content: str
        """

        with open(self._get_path(filename), 'a') as f:
            f.write(content)

    @action
    def getsize(self, filename):
        """
        Get the size of the specified filename in bytes
        """
        return os.path.getsize(filename)

# vim:sw=4:ts=4:et:




from .context import platypush

import unittest

from platypush.event.hook import EventCondition
from platypush.message.event.ping import PingEvent

class TestEventParse(unittest.TestCase):
    def setUp(self):
        self.condition = EventCondition.build({
            'type': 'platypush.message.event.ping.PingEvent',
            'message': 'This is (the)? answer: ${answer}'
        })

    def test_event_parse(self):
        message = "GARBAGE GARBAGE this is the answer: 42"
        event = PingEvent(message=message)
        result = event.matches_condition(self.condition)
        self.assertTrue(result.is_match)
        self.assertTrue('answer' in result.parsed_args)
        self.assertEqual(result.parsed_args['answer'], '42')

        message = "what is not the answer? 43"
        event = PingEvent(message=message)
        result = event.matches_condition(self.condition)
        self.assertFalse(result.is_match)


if __name__ == '__main__':
    unittest.main()

# vim:sw=4:ts=4:et:




import json
import os
import random
import shutil
import threading
from datetime import datetime
from typing import List, Dict, Any, Union, Optional, Tuple, Iterable

import numpy as np
from tensorflow.keras import Model
from tensorflow.keras.layers import Layer
from tensorflow.keras.models import load_model
from tensorflow.keras.preprocessing import image
from tensorflow.keras import utils

from platypush.config import Config
from platypush.context import get_bus
from platypush.message.event.tensorflow import TensorflowEpochStartedEvent, TensorflowEpochEndedEvent, \
    TensorflowBatchStartedEvent, TensorflowBatchEndedEvent, TensorflowTrainStartedEvent, TensorflowTrainEndedEvent
from platypush.message.response.tensorflow import TensorflowTrainResponse, TensorflowPredictResponse
from platypush.plugins import Plugin, action


class TensorflowPlugin(Plugin):
    """
    This plugin can be used to create, train, load and make predictions with TensorFlow-compatible machine learning
    models.

    Triggers:

        - :class:`platypush.message.event.tensorflow.TensorflowEpochStartedEvent`
            when a Tensorflow model training/evaluation epoch begins.
        - :class:`platypush.message.event.tensorflow.TensorflowEpochEndedEvent`
            when a Tensorflow model training/evaluation epoch ends.
        - :class:`platypush.message.event.tensorflow.TensorflowBatchStartedEvent`
            when a Tensorflow model training/evaluation batch starts being processed.
        - :class:`platypush.message.event.tensorflow.TensorflowBatchEndedEvent`
            when a the processing of a Tensorflow model training/evaluation batch ends.
        - :class:`platypush.message.event.tensorflow.TensorflowTrainStartedEvent`
            when a Tensorflow model starts being trained.
        - :class:`platypush.message.event.tensorflow.TensorflowTrainEndedEvent`
            when the training phase of a Tensorflow model ends.

    Requires:

        * **numpy** (``pip install numpy``)
        * **pandas** (``pip install pandas``) (optional, for CSV parsing)
        * **tensorflow** (``pip install 'tensorflow>=2.0'``)
        * **keras** (``pip install keras``)

    """

    _image_extensions = ['jpg', 'jpeg', 'bmp', 'tiff', 'tif', 'png', 'gif']
    _numpy_extensions = ['npy', 'npz']
    _csv_extensions = ['csv', 'tsv']
    _supported_data_file_extensions = [*_csv_extensions, *_numpy_extensions, *_image_extensions]

    def __init__(self, workdir: Optional[str] = None, **kwargs):
        """
        :param workdir: Working directory for TensorFlow, where models will be stored
            (default: PLATYPUSH_WORKDIR/tensorflow).
        """
        super().__init__(**kwargs)
        self.models: Dict[str, Model] = {}
        self._model_locks: Dict[str, threading.RLock()] = {}
        self._work_dir = os.path.abspath(os.path.expanduser(workdir)) if workdir else \
            os.path.join(Config.get('workdir'), 'tensorflow')

        self._models_dir = os.path.join(self._work_dir, 'models')
        os.makedirs(self._models_dir, mode=0o755, exist_ok=True)

    def _load_model(self, name: str, reload: bool = False) -> Model:
        if name in self.models and not reload:
            return self.models[name]

        model_dir = os.path.join(self._models_dir, name)
        assert os.path.isdir(model_dir), 'The model {} does not exist'.format(name)
        model = load_model(model_dir)
        model.input_labels = []
        model.output_labels = []

        labels_file = os.path.join(model_dir, 'labels.json')
        if os.path.isfile(labels_file):
            with open(labels_file, 'r') as f:
                labels = json.load(f)
            if 'input' in labels:
                model.input_labels = labels['input']
            if 'output' in labels:
                model.output_labels = labels['output']

        return model

    def _generate_callbacks(self, model: str):
        from tensorflow.keras.callbacks import LambdaCallback
        return [LambdaCallback(
            on_epoch_begin=self.on_epoch_begin(model),
            on_epoch_end=self.on_epoch_end(model),
            on_batch_begin=self.on_batch_begin(model),
            on_batch_end=self.on_batch_end(model),
            on_train_begin=self.on_train_begin(model),
            on_train_end=self.on_train_end(model),
        )]

    @staticmethod
    def on_epoch_begin(model: str):
        def callback(epoch: int, logs: Optional[dict] = None):
            get_bus().post(TensorflowEpochStartedEvent(model=model, epoch=epoch, logs=logs))
        return callback

    @staticmethod
    def on_epoch_end(model: str):
        def callback(epoch: int, logs: Optional[dict] = None):
            get_bus().post(TensorflowEpochEndedEvent(model=model, epoch=epoch, logs=logs))
        return callback

    @staticmethod
    def on_batch_begin(model: str):
        def callback(batch: int, logs: Optional[dict] = None):
            get_bus().post(TensorflowBatchStartedEvent(model=model, batch=batch, logs=logs))
        return callback

    @staticmethod
    def on_batch_end(model: str):
        def callback(batch, logs: Optional[dict] = None):
            get_bus().post(TensorflowBatchEndedEvent(model=model, batch=batch, logs=logs))
        return callback

    @staticmethod
    def on_train_begin(model: str):
        def callback(logs: Optional[dict] = None):
            get_bus().post(TensorflowTrainStartedEvent(model=model, logs=logs))
        return callback

    @staticmethod
    def on_train_end(model: str):
        def callback(logs: Optional[dict] = None):
            get_bus().post(TensorflowTrainEndedEvent(model=model, logs=logs))
        return callback

    @action
    def load(self, model: str, reload: bool = False) -> Dict[str, Any]:
        """
        (Re)-load a model from the file system.

        :param model: Name of the model. Must be a folder name stored under ``<workdir>/models``.
        :param reload: If ``True``, the model will be reloaded from the filesystem even if it's been already
            loaded, otherwise the model currently in memory will be kept (default: ``False``).
        :return: The model configuration.
        """
        model = self._load_model(model, reload=reload)
        return model.get_config()

    @action
    def unload(self, model: str) -> None:
        """
        Remove a loaded model from memory.

        :param model: Name of the model.
        """
        assert model in self.models, 'The model {} is not loaded'.format(model)
        del self.models[model]

    @action
    def remove(self, model: str) -> None:
        """
        Unload a module and, if stored on the filesystem, remove its resource files as well.
        WARNING: This operation is not reversible.

        :param model: Name of the model.
        """
        if model in self.models:
            del self.models[model]

        model_dir = os.path.join(self._models_dir, model)
        if os.path.isdir(model_dir):
            shutil.rmtree(model_dir)

    @action
    def create_network(self,
                       name: str,
                       layers: List[Union[Layer, Dict[str, Any]]],
                       input_names: Optional[List[str]] = None,
                       output_names: Optional[List[str]] = None,
                       optimizer: Optional[str] = 'rmsprop',
                       loss: Optional[Union[str, List[str], Dict[str, str]]] = None,
                       metrics: Optional[
                           Union[str, List[Union[str, List[str]]], Dict[str, Union[str, List[str]]]]] = None,
                       loss_weights: Optional[Union[List[float], Dict[str, float]]] = None,
                       sample_weight_mode: Optional[Union[str, List[str], Dict[str, str]]] = None,
                       weighted_metrics: Optional[List[str]] = None,
                       target_tensors=None,
                       **kwargs) -> Dict[str, Any]:
        """
        Create a neural network TensorFlow Keras model.

        :param name: Name of the model.
        :param layers: List of layers. Example:

        .. code-block:: javascript

            [
              // Input flatten layer with 10 units
              {
                "type": "Flatten",
                "input_shape": [10, 10]
              },

              // Dense hidden layer with 500 units
              {
                "type": "Dense",
                "units": 500,
                "activation": "relu"
              },

              // Dense hidden layer with 100 units
              {
                "type": "Dense",
                "units": 100,
                "activation": "relu"
              },

              // Dense output layer with 2 units (labels) and ``softmax`` activation function
              {
                "type": "Dense",
                "units": 2,
                "activation": "softmax"
              }
            ]

        :param input_names: List of names for the input units (default: TensorFlow name auto-assign logic).
        :param output_names: List of labels for the output units (default: TensorFlow name auto-assign logic).
        :param optimizer: Optimizer, see <https://keras.io/optimizers/> (default: ``rmsprop``).
        :param loss: Loss function, see <https://keras.io/losses/>. An objective function is any callable with
            the signature ``scalar_loss = fn(y_true, y_pred)``. If the model has multiple outputs, you can use a
            different loss on each output by passing a dictionary or a list of losses. The loss value that will be
            minimized by the model will then be the sum of all individual losses (default: None).

        :param metrics: List of metrics to be evaluated by the model during training and testing. Typically you will
            use ``metrics=['accuracy']``. To specify different metrics for different outputs of a multi-output model,
            you could also pass a dictionary, such as
            ``metrics={'output_a': 'accuracy', 'output_b': ['accuracy', 'mse']}``. You can also pass a list
            ``(len = len(outputs))`` of lists of metrics such as ``metrics=[['accuracy'], ['accuracy', 'mse']]`` or
            ``metrics=['accuracy', ['accuracy', 'mse']]``. Default: ``['accuracy']``.

        :param loss_weights: Optional list or dictionary specifying scalar coefficients (Python floats) to weight the
            loss contributions of different model outputs. The loss value that will be minimized by the model
            will then be the *weighted sum* of all individual losses, weighted by the `loss_weights` coefficients.
            If a list, it is expected to have a 1:1 mapping to the model's outputs. If a tensor, it is expected to map
            output names (strings) to scalar coefficients.

        :param sample_weight_mode: If you need to do time-step-wise sample weighting (2D weights), set this to
            ``"temporal"``. ``None`` defaults to sample-wise weights (1D). If the model has multiple outputs,
            you can use a different ``sample_weight_mode`` on each output by passing a dictionary or a list of modes.

        :param weighted_metrics: List of metrics to be evaluated and weighted by ``sample_weight`` or ``class_weight``
            during training and testing.

        :param target_tensors: By default, Keras will create placeholders for the model's target, which will be fed
            with the target data during training. If instead you would like to use your own target tensors (in turn,
            Keras will not expect external numpy data for these targets at training time), you can specify them via the
            ``target_tensors`` argument. It can be a single tensor (for a single-output model), a list of tensors,
            or a dict mapping output names to target tensors.

        :param kwargs: Extra arguments to pass to ``Model.compile()``.

        :return: The model configuration, as a dict. Example:

        .. code-block:: json

            {
              "name": "test_model",
              "layers": [
                {
                  "class_name": "Flatten",
                  "config": {
                    "name": "flatten",
                    "trainable": true,
                    "batch_input_shape": [
                      null,
                      10
                    ],
                    "dtype": "float32",
                    "data_format": "channels_last"
                  }
                },
                {
                  "class_name": "Dense",
                  "config": {
                    "name": "dense",
                    "trainable": true,
                    "dtype": "float32",
                    "units": 100,
                    "activation": "relu",
                    "use_bias": true,
                    "kernel_initializer": {
                      "class_name": "GlorotUniform",
                      "config": {
                        "seed": null
                      }
                    },
                    "bias_initializer": {
                      "class_name": "Zeros",
                      "config": {}
                    },
                    "kernel_regularizer": null,
                    "bias_regularizer": null,
                    "activity_regularizer": null,
                    "kernel_constraint": null,
                    "bias_constraint": null
                  }
                },
                {
                  "class_name": "Dense",
                  "config": {
                    "name": "dense_1",
                    "trainable": true,
                    "dtype": "float32",
                    "units": 50,
                    "activation": "relu",
                    "use_bias": true,
                    "kernel_initializer": {
                      "class_name": "GlorotUniform",
                      "config": {
                        "seed": null
                      }
                    },
                    "bias_initializer": {
                      "class_name": "Zeros",
                      "config": {}
                    },
                    "kernel_regularizer": null,
                    "bias_regularizer": null,
                    "activity_regularizer": null,
                    "kernel_constraint": null,
                    "bias_constraint": null
                  }
                },
                {
                  "class_name": "Dense",
                  "config": {
                    "name": "dense_2",
                    "trainable": true,
                    "dtype": "float32",
                    "units": 2,
                    "activation": "softmax",
                    "use_bias": true,
                    "kernel_initializer": {
                      "class_name": "GlorotUniform",
                      "config": {
                        "seed": null
                      }
                    },
                    "bias_initializer": {
                      "class_name": "Zeros",
                      "config": {}
                    },
                    "kernel_regularizer": null,
                    "bias_regularizer": null,
                    "activity_regularizer": null,
                    "kernel_constraint": null,
                    "bias_constraint": null
                  }
                }
              ]
            }

        """
        from tensorflow.keras import Sequential
        model = Sequential(name=name)
        for layer in layers:
            if not isinstance(layer, Layer):
                layer = self._layer_from_dict(layer.pop('type'), **layer)
            model.add(layer)

        if not metrics:
            metrics = ['accuracy']

        model.compile(
            optimizer=optimizer,
            loss=loss,
            metrics=metrics,
            loss_weights=loss_weights,
            sample_weight_mode=sample_weight_mode,
            weighted_metrics=weighted_metrics,
            target_tensors=target_tensors,
            **kwargs
        )

        model.input_labels = input_names or []
        model.output_labels = output_names or []
        self.models[name] = model
        return model.get_config()

    @action
    def create_regression(self,
                          name: str,
                          units: int = 1,
                          input_names: Optional[List[str]] = None,
                          output_names: Optional[List[str]] = None,
                          activation: str = 'linear',
                          use_bias: bool = True,
                          kernel_initializer: str = 'glorot_uniform',
                          bias_initializer: str = 'zeros',
                          kernel_regularizer: Optional[str] = None,
                          bias_regularizer: Optional[str] = None,
                          optimizer: Optional[str] = 'rmsprop',
                          loss: Optional[Union[str, List[str], Dict[str, str]]] = 'mse',
                          metrics: Optional[
                              Union[str, List[Union[str, List[str]]], Dict[str, Union[str, List[str]]]]] = None,
                          loss_weights: Optional[Union[List[float], Dict[str, float]]] = None,
                          sample_weight_mode: Optional[Union[str, List[str], Dict[str, str]]] = None,
                          weighted_metrics: Optional[List[str]] = None,
                          target_tensors=None,
                          **kwargs) -> Dict[str, Any]:
        """
        Create a linear/logistic regression model.

        :param name: Name of the model.
        :param units: Output dimension (default: 1).
        :param input_names: List of names for the input units (default: TensorFlow name auto-assign logic).
        :param output_names: List of labels for the output units (default: TensorFlow name auto-assign logic).
        :param activation: Activation function to be used (default: None).
        :param use_bias: Whether to calculate the bias/intercept for this model. If set
            to False, no bias/intercept will be used in calculations, e.g., the data
            is already centered (default: True).
        :param kernel_initializer: Initializer for the ``kernel`` weights matrices (default: ``glorot_uniform``).
        :param bias_initializer: Initializer for the bias vector (default: ``zeros``).
        :param kernel_regularizer: Regularizer for kernel vectors (default: None).
        :param bias_regularizer: Regularizer for bias vectors (default: None).
        :param optimizer: Optimizer, see <https://keras.io/optimizers/> (default: ``rmsprop``).
        :param loss: Loss function, see <https://keras.io/losses/>. An objective function is any callable with
            the signature ``scalar_loss = fn(y_true, y_pred)``. If the model has multiple outputs, you can use a
            different loss on each output by passing a dictionary or a list of losses. The loss value that will be
            minimized by the model will then be the sum of all individual losses (default: ``mse``, mean squared error).

        :param metrics: List of metrics to be evaluated by the model during training and testing. Typically you will
            use ``metrics=['accuracy']``. To specify different metrics for different outputs of a multi-output model,
            you could also pass a dictionary, such as
            ``metrics={'output_a': 'accuracy', 'output_b': ['accuracy', 'mse']}``. You can also pass a list
            ``(len = len(outputs))`` of lists of metrics such as ``metrics=[['accuracy'], ['accuracy', 'mse']]`` or
            ``metrics=['accuracy', ['accuracy', 'mse']]``. Default: ``['mae', 'mse']``.

        :param loss_weights: Optional list or dictionary specifying scalar coefficients (Python floats) to weight the
            loss contributions of different model outputs. The loss value that will be minimized by the model
            will then be the *weighted sum* of all individual losses, weighted by the `loss_weights` coefficients.
            If a list, it is expected to have a 1:1 mapping to the model's outputs. If a tensor, it is expected to map
            output names (strings) to scalar coefficients.

        :param sample_weight_mode: If you need to do time-step-wise sample weighting (2D weights), set this to
            ``"temporal"``. ``None`` defaults to sample-wise weights (1D). If the model has multiple outputs,
            you can use a different ``sample_weight_mode`` on each output by passing a dictionary or a list of modes.

        :param weighted_metrics: List of metrics to be evaluated and weighted by ``sample_weight`` or ``class_weight``
            during training and testing.

        :param target_tensors: By default, Keras will create placeholders for the model's target, which will be fed
            with the target data during training. If instead you would like to use your own target tensors (in turn,
            Keras will not expect external numpy data for these targets at training time), you can specify them via the
            ``target_tensors`` argument. It can be a single tensor (for a single-output model), a list of tensors,
            or a dict mapping output names to target tensors.

        :param kwargs: Extra arguments to pass to ``Model.compile()``.

        :return: Configuration of the model, as a dict. Example:

        .. code-block:: json

            {
              "name": "test_regression_model",
              "trainable": true,
              "dtype": "float32",
              "units": 1,
              "activation": "linear",
              "use_bias": true,
              "kernel_initializer": {
                "class_name": "GlorotUniform",
                "config": {
                  "seed": null
                }
              },
              "bias_initializer": {
                "class_name": "Zeros",
                "config": {}
              },
              "kernel_regularizer": null,
              "bias_regularizer": null
            }

        """
        from tensorflow.keras.experimental import LinearModel
        model = LinearModel(
            units=units,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            name=name)

        model.input_names = input_names or []

        if output_names:
            assert units == len(output_names)
            model.output_labels = output_names
        else:
            model.output_labels = []

        if not metrics:
            metrics = ['mae', 'mse']

        model.compile(
            optimizer=optimizer,
            loss=loss,
            metrics=metrics,
            loss_weights=loss_weights,
            sample_weight_mode=sample_weight_mode,
            weighted_metrics=weighted_metrics,
            target_tensors=target_tensors,
            **kwargs
        )

        self.models[name] = model
        return model.get_config()

    @staticmethod
    def _layer_from_dict(layer_type: str, *args, **kwargs) -> Layer:
        from tensorflow.keras import layers
        cls = getattr(layers, layer_type)
        assert issubclass(cls, Layer)
        return cls(*args, **kwargs)

    @staticmethod
    def _get_csv_data(data_file: str) -> np.ndarray:
        import pandas as pd
        return pd.read_csv(data_file).to_numpy()

    @staticmethod
    def _get_numpy_data(data_file: str) -> np.ndarray:
        return np.load(data_file)

    @staticmethod
    def _get_numpy_compressed_data(data_file: str) -> np.ndarray:
        return list(np.load(data_file).values()).pop()

    @classmethod
    def _get_image(cls, image_file: str, model: Model) -> np.ndarray:
        input_shape = model.inputs[0].shape
        size = input_shape[1:3].as_list()
        assert len(size) == 2, 'The model {} does not have enough dimensions to process an image (shape: {})'.format(
            model.name, size)

        colors = input_shape[3:]
        assert colors, ('The model {} requires a tensor with at least 3 inputs in order to process images: ' +
                        '[WIDTH, HEIGHT, COLORS]').format(model.name)

        if colors[0] == 1:
            color_mode = 'grayscale'
        elif colors[0] == 3:
            color_mode = 'rgb'
        elif colors[0] == 4:
            color_mode = 'rgba'
        else:
            raise AssertionError('The input tensor should have either 1 (grayscale), 3 (rgb) or 4 (rgba) units. ' +
                                 'Found: {}'.format(colors[0]))

        img = image.load_img(image_file, target_size=size, color_mode=color_mode)
        return image.img_to_array(img)

    @classmethod
    def _get_dir(cls, directory: str, model: Model) -> Dict[str, Iterable]:
        labels = [f for f in os.listdir(directory) if os.path.isdir(os.path.join(directory, f))]
        assert set(model.output_labels) == set(labels),\
            'The directory {dir} should contain exactly {n} subfolders named {names}'.format(
                dir=directory, n=len(model.output_labels), names=model.output.labels)

        ret = {}
        for label in labels:
            subdir = os.path.join(directory, label)
            ret[label] = [
                cls._get_data(os.path.join(subdir, f), model)
                for f in os.listdir(subdir)
                if f.split('.')[-1] in cls._supported_data_file_extensions
            ]

        return ret

    @classmethod
    def _get_data(cls, data: Union[str, np.ndarray, Iterable, Dict[str, Union[Iterable, np.ndarray]]], model: Model) \
            -> Union[np.ndarray, Iterable, Dict[str, Union[Iterable, np.ndarray]]]:
        if not isinstance(data, str):
            return data

        if data.startswith('http://') or data.startswith('https://'):
            filename = '{timestamp}_{filename}'.format(
                timestamp=datetime.now().timestamp(), filename=data.split('/')[-1])
            data_file = utils.get_file(filename, data)
        else:
            data_file = os.path.abspath(os.path.expanduser(data))

        extensions = [ext for ext in cls._supported_data_file_extensions if data_file.endswith('.' + ext)]

        if os.path.isfile(data_file):
            assert extensions, 'Unsupported type for file {}. Supported extensions: {}'.format(
                data_file, cls._supported_data_file_extensions
            )

            extension = extensions.pop()
            if extension in cls._csv_extensions:
                return cls._get_csv_data(data_file)
            if extension == 'npy':
                return cls._get_numpy_data(data_file)
            if extension == 'npz':
                return cls._get_numpy_compressed_data(data_file)
            if extension in cls._image_extensions:
                return cls._get_image(data_file, model)

            raise AssertionError('Unsupported file type: {}'.format(data_file))
        elif os.path.isdir(data_file):
            return cls._get_dir(data_file, model)

    @classmethod
    def _get_dataset(cls,
                     inputs: Union[str, np.ndarray, Iterable, Dict[str, Union[Iterable, np.ndarray]]],
                     outputs: Optional[Union[str, np.ndarray, Iterable, Dict[str, Union[Iterable, np.ndarray]]]],
                     model: Model) \
            -> Tuple[Union[np.ndarray, Iterable, Dict[str, Union[Iterable, np.ndarray]]],
                     Optional[Union[np.ndarray, Iterable, Dict[str, Union[Iterable, np.ndarray]]]]]:
        inputs = cls._get_data(inputs, model)
        if outputs:
            outputs = cls._get_data(inputs, model)
        elif isinstance(inputs, dict) and model.output_labels:
            pairs = []
            for i, label in enumerate(model.output_labels):
                data = inputs.get(label, [])
                pairs.extend([(d, i) for d in data])

            random.shuffle(pairs)
            inputs = np.asarray([p[0] for p in pairs])
            outputs = np.asarray([p[1] for p in pairs])

        return inputs, outputs

    @action
    def train(self,
              model: str,
              inputs: Union[str, np.ndarray, Iterable, Dict[str, Union[Iterable, np.ndarray]]],
              outputs: Optional[Union[str, np.ndarray, Iterable]] = None,
              batch_size: Optional[int] = None,
              epochs: int = 1,
              verbose: int = 1,
              validation_split: float = 0.,
              validation_data: Optional[Tuple[Union[np.ndarray, Iterable]]] = None,
              shuffle: Union[bool, str] = True,
              class_weight: Optional[Dict[int, float]] = None,
              sample_weight: Optional[Union[np.ndarray, Iterable]] = None,
              initial_epoch: int = 0,
              steps_per_epoch: Optional[int] = None,
              validation_steps: int = None,
              validation_freq: int = 1,
              max_queue_size: int = 10,
              workers: int = 1,
              use_multiprocessing: bool = False) -> TensorflowTrainResponse:
        """
        Trains a model on a dataset for a fixed number of epochs.

        :param model: Name of the existing model to be trained.
        :param inputs: Input data. It can be:

          - A numpy array (or array-like), or a list of arrays in case the model has multiple inputs.
          - A TensorFlow tensor, or a list of tensors in case the model has multiple inputs.
          - A dict mapping input names to the corresponding array/tensors, if the model has named inputs.
          - A ``tf.data`` dataset. Should return a tuple of either ``(inputs, targets)`` or
            ``(inputs, targets, sample_weights)``.
          - A generator or ``keras.utils.Sequence`` returning ``(inputs, targets)`` or
            ``(inputs, targets, sample weights)``.
          - A string that points to a file. Supported formats:

              - CSV with header (``.csv`` extension``)
              - Numpy raw or compressed files (``.npy`` or ``.npz`` extension)
              - Image files
              - An HTTP URL pointing to one of the file types listed above
              - Directories with images. If ``inputs`` points to a directory of images then the following
                conventions are followed:

                - The folder must contain exactly as many subfolders as the output units of your model. If
                  the model has ``output_labels`` then those subfolders should be named as the output labels.
                  Each subfolder will contain training examples that match the associated label (e.g.
                  ``positive`` will contain all the positive images and ``negative`` all the negative images).
                - ``outputs`` doesn't have to be specified.

        :param outputs: Target data. Like the input data `x`, it can be a numpy array (or array-like) or TensorFlow
            tensor(s). It should be consistent with `x` (you cannot have Numpy inputs and tensor targets, or inversely).
            If `x` is a dataset, generator, or `keras.utils.Sequence` instance, `y` should not be specified
            (since targets will be obtained from `x`).

        :param batch_size: Number of samples per gradient update. If unspecified, ``batch_size`` will default to 32.
            Do not specify the ``batch_size`` if your data is in the form of symbolic tensors, datasets,
            generators, or ``keras.utils.Sequence`` instances (since they generate batches).

        :param epochs: Number of epochs to train the model. An epoch is an iteration over the entire ``x`` and ``y``
            data provided. Note that in conjunction with ``initial_epoch``, ``epochs`` is to be understood as
            "final epoch". The model is not trained for a number of iterations given by ``epochs``, but merely until
            the epoch of index ``epochs`` is reached.

        :param verbose: Verbosity mode. 0 = silent, 1 = progress bar, 2 = one line per epoch.
            Note that the progress bar is not particularly useful when
            logged to a file, so verbose=2 is recommended when not running
            interactively (eg, in a production environment).

        :param validation_split: Float between 0 and 1.
            Fraction of the training data to be used as validation data. The model will set apart this fraction
            of the training data, will not train on it, and will evaluate the loss and any model metrics on this data
            at the end of each epoch. The validation data is selected from the last samples in the ``x`` and ``y``
            data provided, before shuffling. This argument is not supported when ``x`` is a dataset, generator or
           ``keras.utils.Sequence`` instance.

        :param validation_data: Data on which to evaluate the loss and any model metrics at the end of each epoch.
            The model will not be trained on this data. ``validation_data`` will override ``validation_split``.
            ``validation_data`` could be:

              - tuple ``(x_val, y_val)`` of arrays/numpy arrays/tensors
              - tuple ``(x_val, y_val, val_sample_weights)`` of Numpy arrays
              - dataset

            For the first two cases, ``batch_size`` must be provided. For the last case, ``validation_steps`` could be
            provided.

        :param shuffle: Boolean (whether to shuffle the training data before each epoch) or str (for 'batch').
            'batch' is a special option for dealing with the limitations of HDF5 data; it shuffles in batch-sized
            chunks. Has no effect when ``steps_per_epoch`` is not ``None``.

        :param class_weight: Optional dictionary mapping class indices (integers) to a weight (float) value, used
            for weighting the loss function (during training only). This can be useful to tell the model to
            "pay more attention" to samples from an under-represented class.

        :param sample_weight: Optional iterable/numpy array of weights for the training samples, used for weighting
            the loss function (during training only). You can either pass a flat (1D) numpy array/iterable with the
            same length as the input samples (1:1 mapping between weights and samples), or in the case of temporal data,
            you can pass a 2D array with shape ``(samples, sequence_length)``, to apply a different weight to every
            time step of every sample. In this case you should make sure to specify ``sample_weight_mode="temporal"``
            in ``compile()``. This argument is not supported when ``x`` is a dataset, generator, or
            ``keras.utils.Sequence`` instance, instead provide the sample_weights as the third element of ``x``.

        :param initial_epoch: Epoch at which to start training (useful for resuming a previous training run).

        :param steps_per_epoch: Total number of steps (batches of samples) before declaring one epoch finished and
            starting the next epoch. When training with input tensors such as TensorFlow data tensors, the default
            ``None`` is equal to the number of samples in your dataset divided by the batch size, or 1 if that cannot
            be determined. If x is a ``tf.data`` dataset, and 'steps_per_epoch' is None, the epoch will run until the
            input dataset is exhausted. This argument is not supported with array inputs.

        :param validation_steps: Only relevant if ``validation_data`` is provided and is a ``tf.data`` dataset. Total
            number of steps (batches of samples) to draw before stopping when performing validation at the end of
            every epoch. If 'validation_steps' is None, validation will run until the ``validation_data`` dataset is
            exhausted. In the case of a infinite dataset, it will run into a infinite loop. If 'validation_steps' is
            specified and only part of the dataset will be consumed, the evaluation will start from the beginning of
            the dataset at each epoch. This ensures that the same validation samples are used every time.

        :param validation_freq: Only relevant if validation data is provided. Integer or ``collections_abc.Container``
            instance (e.g. list, tuple, etc.). If an integer, specifies how many training epochs to run before a
            new validation run is performed, e.g. ``validation_freq=2`` runs validation every 2 epochs. If a
            Container, specifies the epochs on which to run validation, e.g. ``validation_freq=[1, 2, 10]`` runs
            validation at the end of the 1st, 2nd, and 10th epochs.

        :param max_queue_size: Used for generator or ``keras.utils.Sequence`` input only. Maximum size for
            the generator queue. If unspecified, ``max_queue_size`` will default to 10.

        :param workers: Used for generator or ``keras.utils.Sequence`` input only. Maximum number of processes
            to spin up when using process-based threading. If unspecified, ``workers`` will default to 1. If 0, will
            execute the generator on the main thread.

        :param use_multiprocessing: Used for generator or ``keras.utils.Sequence`` input only. If ``True``,
            use process-based threading. If unspecified, ``use_multiprocessing`` will default to ``False``.
            Note that because this implementation relies on multiprocessing, you should not pass non-picklable
            arguments to the generator as they can't be passed easily to children processes.

        :return: :class:`platypush.message.response.tensorflow.TensorflowTrainResponse`
        """
        name = model
        model = self._load_model(model)
        inputs, outputs = self._get_dataset(inputs, outputs, model)

        ret = model.fit(
            x=inputs,
            y=outputs,
            batch_size=batch_size,
            epochs=epochs,
            verbose=verbose,
            callbacks=self._generate_callbacks(name),
            validation_split=validation_split,
            validation_data=validation_data,
            shuffle=shuffle,
            class_weight=class_weight,
            sample_weight=sample_weight,
            initial_epoch=initial_epoch,
            steps_per_epoch=steps_per_epoch,
            validation_steps=validation_steps,
            validation_freq=validation_freq,
            max_queue_size=max_queue_size,
            workers=workers,
            use_multiprocessing=use_multiprocessing,
        )

        return TensorflowTrainResponse(model=model, epochs=ret.epoch, history=ret.history)

    @action
    def evaluate(self,
                 model: str,
                 inputs: Union[str, np.ndarray, Iterable, Dict[str, Union[Iterable, np.ndarray]]],
                 outputs: Optional[Union[str, np.ndarray, Iterable]] = None,
                 batch_size: Optional[int] = None,
                 verbose: int = 1,
                 sample_weight: Optional[Union[np.ndarray, Iterable]] = None,
                 steps: Optional[int] = None,
                 max_queue_size: int = 10,
                 workers: int = 1,
                 use_multiprocessing: bool = False) -> Union[Dict[str, float], List[float]]:
        """
        Returns the loss value and metrics values for the model in test model.

        :param model: Name of the existing model to be trained.
        :param inputs: Input data. It can be:

          - A numpy array (or array-like), or a list of arrays in case the model has multiple inputs.
          - A TensorFlow tensor, or a list of tensors in case the model has multiple inputs.
          - A dict mapping input names to the corresponding array/tensors, if the model has named inputs.
          - A ``tf.data`` dataset. Should return a tuple of either ``(inputs, targets)`` or
            ``(inputs, targets, sample_weights)``.
          - A generator or ``keras.utils.Sequence`` returning ``(inputs, targets)`` or
            ``(inputs, targets, sample weights)``.
          - A string that points to a file. Supported formats:

              - CSV with header (``.csv`` extension``)
              - Numpy raw or compressed files (``.npy`` or ``.npz`` extension)
              - Image files
              - An HTTP URL pointing to one of the file types listed above
              - Directories with images. If ``inputs`` points to a directory of images then the following
                conventions are followed:

                - The folder must contain exactly as many subfolders as the output units of your model. If
                  the model has ``output_labels`` then those subfolders should be named as the output labels.
                  Each subfolder will contain training examples that match the associated label (e.g.
                  ``positive`` will contain all the positive images and ``negative`` all the negative images).
                - ``outputs`` doesn't have to be specified.


        :param outputs: Target data. Like the input data `x`, it can be a numpy array (or array-like) or TensorFlow tensor(s).
            It should be consistent with `x` (you cannot have Numpy inputs and tensor targets, or inversely).
            If `x` is a dataset, generator, or `keras.utils.Sequence` instance, `y` should not be specified
            (since targets will be obtained from `x`).

        :param batch_size: Number of samples per gradient update. If unspecified, ``batch_size`` will default to 32.
            Do not specify the ``batch_size`` if your data is in the form of symbolic tensors, datasets,
            generators, or ``keras.utils.Sequence`` instances (since they generate batches).

        :param verbose: Verbosity mode. 0 = silent, 1 = progress bar, 2 = one line per epoch.
            Note that the progress bar is not particularly useful when
            logged to a file, so verbose=2 is recommended when not running
            interactively (eg, in a production environment).

        :param sample_weight: Optional iterable/numpy array of weights for the training samples, used for weighting
            the loss function (during training only). You can either pass a flat (1D) numpy array/iterable with the
            same length as the input samples (1:1 mapping between weights and samples), or in the case of temporal data,
            you can pass a 2D array with shape ``(samples, sequence_length)``, to apply a different weight to every
            time step of every sample. In this case you should make sure to specify ``sample_weight_mode="temporal"``
            in ``compile()``. This argument is not supported when ``x`` is a dataset, generator, or
            ``keras.utils.Sequence`` instance, instead provide the sample_weights as the third element of ``x``.

        :param steps: Total number of steps (batches of samples) before declaring the evaluation round finished.
            Ignored with the default value of ``None``. If x is a ``tf.data`` dataset and ``steps`` is None, 'evaluate'
            will run until the dataset is exhausted. This argument is not supported with array inputs.

        :param max_queue_size: Used for generator or ``keras.utils.Sequence`` input only. Maximum size for the generator
            queue. If unspecified, ``max_queue_size`` will default to 10.

        :param workers: Used for generator or ``keras.utils.Sequence`` input only. Maximum number of processes
            to spin up when using process-based threading. If unspecified, ``workers`` will default to 1. If 0, will
            execute the generator on the main thread.

        :param use_multiprocessing: Used for generator or ``keras.utils.Sequence`` input only. If ``True``,
            use process-based threading. If unspecified, ``use_multiprocessing`` will default to ``False``.
            Note that because this implementation relies on multiprocessing, you should not pass non-picklable
            arguments to the generator as they can't be passed easily to children processes.

        :return: ``{test_metric: metric_value}`` dictionary if the ``metrics_names`` of the model are specified,
            otherwise a list with the result test metrics (loss is usually the first value).
        """

        name = model
        model = self._load_model(model)
        inputs, outputs = self._get_dataset(inputs, outputs, model)

        ret = model.evaluate(
            x=inputs,
            y=outputs,
            batch_size=batch_size,
            verbose=verbose,
            sample_weight=sample_weight,
            steps=steps,
            callbacks=self._generate_callbacks(name),
            max_queue_size=max_queue_size,
            workers=workers,
            use_multiprocessing=use_multiprocessing
        )

        ret = ret if isinstance(ret, list) else [ret]
        if not model.metrics_names:
            return ret

        return {model.metrics_names[i]: value for i, value in enumerate(ret)}

    @action
    def predict(self,
                model: str,
                inputs: Union[str, np.ndarray, Iterable, Dict[str, Union[Iterable, np.ndarray]]],
                batch_size: Optional[int] = None,
                verbose: int = 0,
                steps: Optional[int] = None,
                max_queue_size: int = 10,
                workers: int = 1,
                use_multiprocessing: bool = False) -> TensorflowPredictResponse:
        """
        Generates output predictions for the input samples.

        :param model: Name of the existing model to be trained.
        :param inputs: Input data. It can be:

          - A numpy array (or array-like), or a list of arrays in case the model has multiple inputs.
          - A TensorFlow tensor, or a list of tensors in case the model has multiple inputs.
          - A dict mapping input names to the corresponding array/tensors, if the model has named inputs.
          - A ``tf.data`` dataset. Should return a tuple of either ``(inputs, targets)`` or
            ``(inputs, targets, sample_weights)``.
          - A generator or ``keras.utils.Sequence`` returning ``(inputs, targets)`` or
            ``(inputs, targets, sample weights)``.
          - A string that points to a file. Supported formats:

              - CSV with header (``.csv`` extension``)
              - Numpy raw or compressed files (``.npy`` or ``.npz`` extension)
              - Image files
              - An HTTP URL pointing to one of the file types listed above

        :param batch_size: Number of samples per gradient update. If unspecified, ``batch_size`` will default to 32.
            Do not specify the ``batch_size`` if your data is in the form of symbolic tensors, datasets,
            generators, or ``keras.utils.Sequence`` instances (since they generate batches).

        :param verbose: Verbosity mode, 0 or 1.

        :param steps: Total number of steps (batches of samples) before declaring the prediction round finished.
            Ignored with the default value of ``None``. If x is a ``tf.data`` dataset and ``steps`` is None, ``predict``
            will run until the input dataset is exhausted.

        :param max_queue_size: Integer. Used for generator or ``keras.utils.Sequence`` input only. Maximum size for
            the generator queue (default: 10).

        :param workers: Used for generator or ``keras.utils.Sequence`` input only. Maximum number of processes
            to spin up when using process-based threading. If unspecified, ``workers`` will default to 1. If 0, will
            execute the generator on the main thread.

        :param use_multiprocessing: Used for generator or ``keras.utils.Sequence`` input only. If ``True``,
            use process-based threading. If unspecified, ``use_multiprocessing`` will default to ``False``.
            Note that because this implementation relies on multiprocessing, you should not pass non-picklable
            arguments to the generator as they can't be passed easily to children processes.

        :return: :class:`platypush.message.response.tensorflow.TensorflowPredictResponse`. Format:

            - For regression models with no output labels specified: ``values`` will contain the output vector:

            .. code-block:: json

                {
                    "values": [[3.1415]]
                }

            - For regression models with output labels specified: ``values`` will be a list of ``{label -> value}``
              maps:

            .. code-block:: json

                {
                    "values": [
                        {
                            "x": 42.0,
                            "y": 43.0
                        }
                    ]
                }

            - For neural networks: ``values`` will contain the list of the output vector like in the case of
              regression, and ``labels`` will store the list of ``argmax`` (i.e. the index of the output unit with the
              highest value) or their labels, if the model has output labels:

            .. code-block:: json

                {
                    "labels": [
                        "positive"
                    ],
                    "values": [
                        {
                            "positive": 0.998,
                            "negative": 0.002
                        }
                    ]
                }

        """
        name = model
        model = self._load_model(model)
        inputs = self._get_data(inputs, model)
        if isinstance(inputs, np.ndarray) and \
                len(model.inputs[0].shape) == len(inputs.shape) + 1 and \
                model.inputs[0].shape[0] is None:
            inputs = np.asarray([inputs])

        ret = model.predict(
            inputs,
            batch_size=batch_size,
            verbose=verbose,
            steps=steps,
            callbacks=self._generate_callbacks(name),
            max_queue_size=max_queue_size,
            workers=workers,
            use_multiprocessing=use_multiprocessing
        )

        return TensorflowPredictResponse(model=model, prediction=ret, output_labels=model.output_labels)

    @action
    def save(self, model: str, overwrite: bool = True, **opts) -> None:
        """
        Save a model in memory to the filesystem. The model files will be stored under
        ``<WORKDIR>/models/<model_name>``.

        :param model: Model name.
        :param overwrite: Overwrite the model files if they already exist.
        :param opts: Extra options to be passed to ``Model.save()``.
        """
        assert model in self.models, 'No such model in memory: {}'.format(model)
        model_dir = os.path.join(self._models_dir, model)
        model = self.models[model]
        os.makedirs(model_dir, exist_ok=True)
        labels = {}
        labels_file = os.path.join(model_dir, 'labels.json')

        if model.input_labels:
            labels['input'] = model.input_labels
        if model.output_labels:
            labels['output'] = model.output_labels
        if labels:
            with open(labels_file, 'w') as f:
                json.dump(labels, f)

        model.save(model_dir, overwrite=overwrite, options=opts)


# vim:sw=4:ts=4:et:



import json

from flask import abort, jsonify, request, Blueprint

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import logger, get_remote_base_url
from platypush.backend.http.app.routes.plugins.media import media_map, \
    stream_media, register_media, unregister_media

media = Blueprint('media', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    media,
]

@media.route('/media', methods=['GET'])
def get_media():
    """
    This route can be used to get the list of registered streams
    """
    return jsonify([dict(media) for media in media_map.values()])


@media.route('/media', methods=['PUT'])
def add_media():
    """
    This route can be used by the `media` plugin to add streaming content over HTTP
    """

    args = {}
    try:
        args = json.loads(request.data.decode('utf-8'))
    except:
        abort(400, 'Invalid JSON request')

    source = args.get('source')
    if not source:
        abort(400, 'The request does not contain any source')

    subtitles = args.get('subtitles')
    try:
        media_hndl = register_media(source, subtitles)
        ret = dict(media_hndl)
        if media_hndl.subtitles:
            ret['subtitles_url'] = get_remote_base_url() + \
                '/media/subtitles/' + media_hndl.media_id + '.vtt'
        return jsonify(ret)
    except FileNotFoundError as e:
        abort(404, str(e))
    except AttributeError as e:
        abort(400, str(e))
    except Exception as e:
        logger().exception(e)
        abort(500, str(e))


@media.route('/media/<media_id>', methods=['GET', 'DELETE'])
def stream_or_delete_media(media_id):
    """
    This route can be used to stream active media points or unregister
    a mounted media stream
    """

    # Remove the extension
    media_id = '.'.join(media_id.split('.')[:-1])

    try:
        if request.method == 'GET':
            if media_id is None:
                return jsonify([dict(media) for media in media_map.values()])
            else:
                return stream_media(media_id, request)
        else:
            media_info = unregister_media(media_id)
            return jsonify(media_info)
    except (AttributeError, FileNotFoundError) as e:
        abort(404, str(e))
    except KeyError as e:
        abort(400, str(e))
    except Exception as e:
        logger().exception(e)
        abort(500, str(e))


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class JoystickEvent(Event):
    """
    Event triggered upon joystick event
    """

    def __init__(self, code, state, *args, **kwargs):
        """
        :param code: Event code, usually the code of the source key/handle
        :type code: str

        :param state: State of the triggering element. Can be 0/1 for a button, -1/0/1 for an axis, a discrete integer for an analog input etc.
        :type state: int
        """

        super().__init__(*args, code=code, state=state, **kwargs)


# vim:sw=4:ts=4:et:




import json
import os

from platypush.plugins import Plugin, action


class GooglePubsubPlugin(Plugin):
    """
    Send messages over a Google pub/sub instance.
    You'll need a Google Cloud active project and a set of credentials to use this plugin:

        1. Create a project on the `Google Cloud console <https://console.cloud.google.com/projectcreate>`_ if
           you don't have one already.

        2. In the `Google Cloud API console <https://console.cloud.google.com/apis/credentials/serviceaccountkey>`_
           create a new service account key. Select "New Service Account", choose the role "Pub/Sub Editor" and leave
           the key type as JSON.

        3. Download the JSON service credentials file. By default platypush will look for the credentials file under
           ~/.credentials/platypush/google/pubsub.json.

    Requires:

        * **google-cloud-pubsub** (``pip install google-cloud-pubsub``)

    """

    publisher_audience = 'https://pubsub.googleapis.com/google.pubsub.v1.Publisher'
    subscriber_audience = 'https://pubsub.googleapis.com/google.pubsub.v1.Subscriber'
    default_credentials_file = os.path.join(os.path.expanduser('~'),
                                            '.credentials', 'platypush', 'google', 'pubsub.json')

    def __init__(self, credentials_file: str = default_credentials_file, **kwargs):
        """
        :param credentials_file: Path to the JSON credentials file for Google pub/sub (default:
            ~/.credentials/platypush/google/pubsub.json)
        """
        super().__init__(**kwargs)
        self.credentials_file = credentials_file
        self.project_id = self.get_project_id()

    def get_project_id(self):
        credentials = json.load(open(self.credentials_file))
        return credentials.get('project_id')

    def get_credentials(self, audience: str):
        # noinspection PyPackageRequirements
        from google.auth import jwt
        return jwt.Credentials.from_service_account_file(self.credentials_file, audience=audience)

    @action
    def send_message(self, topic: str, msg, **kwargs):
        """
        Sends a message to a topic

        :param topic: Topic/channel where the message will be delivered. You can either specify the full topic name in
            the format ``projects/<project_id>/topics/<topic_name>``, where ``<project_id>`` must be the ID of your
            Google Pub/Sub project, or just ``<topic_name>``  - in such case it's implied that you refer to the
            ``topic_name`` under the ``project_id`` of your service credentials.
        :param msg: Message to be sent. It can be a list, a dict, or a Message object
        :param kwargs: Extra arguments to be passed to .publish()
        """
        # noinspection PyPackageRequirements
        from google.cloud import pubsub_v1
        # noinspection PyPackageRequirements
        from google.api_core.exceptions import AlreadyExists

        credentials = self.get_credentials(self.publisher_audience)
        publisher = pubsub_v1.PublisherClient(credentials=credentials)

        if not topic.startswith('projects/{}/topics/'.format(self.project_id)):
            topic = 'projects/{}/topics/{}'.format(self.project_id, topic)

        try:
            publisher.create_topic(topic)
        except AlreadyExists:
            pass

        if isinstance(msg, int) or isinstance(msg, float):
            msg = str(msg)
        if isinstance(msg, dict) or isinstance(msg, list):
            msg = json.dumps(msg)
        if isinstance(msg, str):
            msg = msg.encode()

        publisher.publish(topic, msg, **kwargs)


# vim:sw=4:ts=4:et:



from typing import Dict, Any

from platypush.message.event import Event


class FoursquareCheckinEvent(Event):
    """
    Event triggered when a new check-in occurs.
    """
    def __init__(self, checkin: Dict[str, Any], *args, **kwargs):
        super().__init__(*args, checkin=checkin, **kwargs)


# vim:sw=4:ts=4:et:



import asyncio
import json
import os
import websockets

from platypush.context import get_or_create_event_loop
from platypush.message import Message
from platypush.plugins import Plugin, action
from platypush.utils import get_ssl_client_context


class WebsocketPlugin(Plugin):
    """
    Plugin to send messages over a websocket connection

    Requires:

        * **websockets** (``pip install websockets``)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @action
    def send(self, url, msg, ssl_cert=None, ssl_key=None, ssl_cafile=None,
             ssl_capath=None, *args, **kwargs):
        """
        Sends a message to a websocket.

        :param url: Websocket URL, e.g. ws://localhost:8765 or wss://localhost:8765
        :type topic: str

        :param msg: Message to be sent. It can be a list, a dict, or a Message object

        :param ssl_cert: Path to the SSL certificate to be used, if the SSL connection requires client authentication as well (default: None)
        :type ssl_cert: str

        :param ssl_key: Path to the SSL key to be used, if the SSL connection requires client authentication as well (default: None)
        :type ssl_key: str

        :param ssl_cafile: Path to the certificate authority file if required by the SSL configuration (default: None)
        :type ssl_cafile: str

        :param ssl_capath: Path to the certificate authority directory if required by the SSL configuration (default: None)
        :type ssl_capath: str
        """

        async def send():
            websocket_args = {}
            if ssl_cert:
                websocket_args['ssl'] = get_ssl_client_context(ssl_cert=ssl_cert,
                                                               ssl_key=ssl_key,
                                                               ssl_cafile=ssl_cafile,
                                                               ssl_capath=ssl_capath)

            async with websockets.connect(url, **websocket_args) as websocket:
                try:
                    await websocket.send(str(msg))
                except websockets.exceptions.ConnectionClosed:
                    self.logger.warning('Error on websocket {}: {}'.
                                        format(url, e))

        try: msg = json.dumps(msg)
        except: pass

        try: msg = Message.build(json.loads(msg))
        except: pass

        loop = get_or_create_event_loop()
        loop.run_until_complete(send())


# vim:sw=4:ts=4:et:




from typing import List

from platypush.message.event import Event
from platypush.message.response.qrcode import ResultModel


class QrcodeEvent(Event):
    pass


class QrcodeScannedEvent(Event):
    """
    Event triggered when a QR-code or bar code is scanned.
    """
    def __init__(self, results: List[ResultModel], *args, **kwargs):
        super().__init__(*args, results=results, **kwargs)


# vim:sw=4:ts=4:et:



import enum
import time

from typing import Optional, Dict, Union, Callable, Tuple, Type

from pyfirmata2 import Arduino, ArduinoMega, ArduinoDue, ArduinoNano, Pin, util, ANALOG, INPUT, PWM

from platypush.plugins import action
from platypush.plugins.sensor import SensorPlugin


class PinType(enum.IntEnum):
    ANALOG = 1
    DIGITAL = 2


class BoardType(enum.Enum):
    MEGA = 'mega'
    DUE = 'due'
    NANO = 'nano'


# noinspection PyBroadException
class ArduinoPlugin(SensorPlugin):
    """
    Interact with an Arduino connected to the host machine over USB using the
    `Firmata <https://www.arduino.cc/en/reference/firmata>`_ protocol.

    You have two options to communicate with an Arduino-compatible board over USB:

        - Use this plugin if you want to use the general-purpose Firmata protocol - in this case most of your
          processing logic will be on the host side and you can read/write data to the Arduino transparently.
        - Use the :class:`platypush.plugins.serial.SerialPlugin` if instead you want to run more custom logic
          on the Arduino and communicate back with the host computer through JSON formatted messages.

    Download and flash the
    `Standard Firmata <https://github.com/firmata/arduino/blob/master/examples/StandardFirmata/StandardFirmata.ino>`_
    firmware to the Arduino in order to use this plugin.

    Requires:

        * **pyfirmata2** (``pip install pyfirmata2``)

    """

    def __init__(self,
                 board: Optional[str] = None,
                 board_type: Optional[str] = None,
                 baud_rate: int = 57600,
                 analog_pins: Optional[Dict[str, int]] = None,
                 digital_pins: Optional[Dict[str, int]] = None,
                 timeout: float = 20.0,
                 conv_functions: Optional[Dict[Union[str, int], Union[str, Callable]]] = None,
                 **kwargs):
        """
        :param board: Default board name or path (e.g. ``COM3`` on Windows or ``/dev/ttyUSB0`` on Unix). If not set
            then the plugin will attempt an auto-discovery.

        :param board_type: Default board type. It can be 'mega', 'due' or 'nano'. Leave empty for auto-detection.
        :param baud_rate: Default serial baud rate (default: 57600)
        :param analog_pins: Optional analog PINs map name->pin_number.
        :param digital_pins: Optional digital PINs map name->pin_number.
        :param timeout: Board communication timeout in seconds.
        :param conv_functions: Optional mapping of conversion functions to apply to the analog values read from a
            certain PIN. The key can either be the PIN number or the name as specified in ``analog_pins``, the value
            can be a function that takes an argument and transforms it or its lambda string representation.
            Note that ``analog_read`` returns by default float values in the range [0.0, 1.0]. Example:

            .. code-block:: yaml

                arduino:
                    board: /dev/ttyUSB0
                    analog_pins:
                        temperature: 1  # Analog PIN 1

                    conv_functions:
                        temperature: 'lambda t: t * 500.0'

        """
        super().__init__(**kwargs)

        self.board = board
        self.board_type = self._get_board_type(board_type)
        self.baud_rate = baud_rate
        self.timeout = timeout

        self._pin_number_by_name = {
            PinType.ANALOG: analog_pins or {},
            PinType.DIGITAL: digital_pins or {},
        }

        self._pin_name_by_number = {
            PinType.ANALOG: {number: name for name, number in self._pin_number_by_name[PinType.ANALOG].items()},
            PinType.DIGITAL: {number: name for name, number in self._pin_number_by_name[PinType.DIGITAL].items()},
        }

        self.conv_functions: Dict[Union[str, int], Callable] = {
            (self._pin_number_by_name[PinType.ANALOG].get(pin, pin)): (f if callable(f) else eval(f))
            for pin, f in (conv_functions or {}).items()
        }

        self._boards = {}
        self._board_iterators = {}

    @staticmethod
    def _get_board_type(board_type: Optional[str] = None) -> Type:
        if not board_type:
            return Arduino

        board_type = board_type.lower()
        if board_type == BoardType.DUE.value:
            return ArduinoDue
        if board_type == BoardType.NANO.value:
            return ArduinoNano
        if board_type == BoardType.MEGA.value:
            return ArduinoMega

        raise AssertionError('Invalid board_type: {}'.format(board_type))

    def _get_board(self,
                   board: Optional[str] = None,
                   board_type: Optional[str] = None,
                   baud_rate: Optional[int] = None,
                   timeout: Optional[float] = None):
        board_name = board or self.board or Arduino.AUTODETECT
        baud_rate = baud_rate or self.baud_rate
        timeout = timeout or self.timeout

        if board_name in self._boards:
            return self._boards[board_name]

        board_type = self._get_board_type(board_type) if board_type else self.board_type
        board = board_type(board_name, baudrate=baud_rate, timeout=timeout)
        self.logger.info('Connected to board {}'.format(board.name))
        self._boards[board_name] = board

        self._board_iterators[board.name] = util.Iterator(board)
        self._board_iterators[board.name].start()
        return board

    def _get_board_and_pin(self,
                           pin: Union[int, str],
                           pin_type: PinType,
                           board: Optional[str] = None,
                           board_type: Optional[str] = None,
                           baud_rate: Optional[int] = None,
                           timeout: Optional[int] = None) -> Tuple[Arduino, int]:
        board = self._get_board(board, board_type=board_type, baud_rate=baud_rate, timeout=timeout)
        if pin in self._pin_number_by_name[pin_type]:
            pin = self._pin_number_by_name[pin_type][pin]

        assert isinstance(pin, int), 'Invalid PIN number/name: {}'.format(pin)
        return board, pin

    @staticmethod
    def _get_pin(pin: int, board: Arduino, pin_type: PinType) -> Pin:
        pins = None
        if pin_type == PinType.ANALOG:
            pins = board.analog
        if pin_type == PinType.DIGITAL:
            pins = board.digital

        assert pins, 'Invalid pin_type: {}'.format(pin_type)

        if pins[pin].mode in [ANALOG, INPUT]:
            pins[pin].enable_reporting()
        return pins[pin]

    def _poll_value(self,
                    pin: int,
                    board: Arduino,
                    pin_type: PinType,
                    timeout: Optional[float] = None) -> Optional[Union[bool, float]]:
        value = None
        poll_start = time.time()

        while value is None:
            if timeout and time.time() - poll_start >= timeout:
                raise RuntimeError('Read timeout')

            pin_ = self._get_pin(pin=pin, board=board, pin_type=pin_type)
            if pin_.mode not in [INPUT, ANALOG]:
                self.logger.warning('PIN {} is not configured in input/analog mode'.format(pin))
                return

            value = pin_.read()
            if value is None:
                time.sleep(0.001)

        if pin_type == PinType.DIGITAL:
            value = bool(value)

        return value

    @action
    def analog_read(self,
                    pin: Union[int, str],
                    board: Optional[str] = None,
                    board_type: Optional[str] = None,
                    baud_rate: Optional[int] = None,
                    conv_function: Optional[Union[str, Callable]] = None,
                    timeout: Optional[int] = None) -> float:
        """
        Read an analog value from a PIN.

        :param pin: PIN number or configured name.
        :param board: Board path or name (default: default configured ``board``).
        :param board_type: Board type. It can be 'mega', 'due' or 'nano' (default: configured ``board_type``).
        :param baud_rate: Baud rate (default: default configured ``baud_rate``).
        :param conv_function: Optional conversion function override to apply to the output. It can be either a function
            object or its lambda string representation (e.g. ``lambda x: x*x``). Keep in mind that ``analog_read``
            returns by default float values in the range ``[0.0, 1.0]``.
        :param timeout: Communication timeout in seconds (default: default configured ``timeout``).
        """
        conv_function = conv_function or self.conv_functions.get(pin)
        board, pin = self._get_board_and_pin(pin=pin,
                                             pin_type=PinType.ANALOG,
                                             board=board,
                                             board_type=board_type,
                                             baud_rate=baud_rate,
                                             timeout=timeout)

        conv_function = conv_function or self.conv_functions.get(pin)
        value = self._poll_value(pin=pin, board=board, pin_type=PinType.ANALOG, timeout=timeout)

        if conv_function:
            value = conv_function(value)
        return value

    @action
    def digital_read(self,
                     pin: Union[int, str],
                     board: Optional[str] = None,
                     board_type: Optional[str] = None,
                     baud_rate: Optional[int] = None,
                     timeout: Optional[int] = None) -> bool:
        """
        Read a digital value from a PIN.

        :param pin: PIN number or configured name.
        :param board: Board path or name (default: default configured ``board``).
        :param board_type: Board type. It can be 'mega', 'due' or 'nano' (default: configured ``board_type``).
        :param baud_rate: Baud rate (default: default configured ``baud_rate``).
        :param timeout: Communication timeout in seconds (default: default configured ``timeout``).
        """
        board, pin = self._get_board_and_pin(pin=pin,
                                             pin_type=PinType.DIGITAL,
                                             board=board,
                                             board_type=board_type,
                                             baud_rate=baud_rate,
                                             timeout=timeout)

        return self._poll_value(pin=pin, board=board, pin_type=PinType.DIGITAL, timeout=timeout)

    @action
    def analog_write(self, pin: Union[int, str],
                     value: float,
                     board: Optional[str] = None,
                     board_type: Optional[str] = None,
                     baud_rate: Optional[int] = None,
                     timeout: Optional[int] = None):
        """
        Write a value to an analog PIN.

        :param pin: PIN number or configured name.
        :param value: Voltage to be sent, a real number normalized between 0 and 1.
        :param board: Board path or name (default: default configured ``board``).
        :param board_type: Board type. It can be 'mega', 'due' or 'nano' (default: configured ``board_type``).
        :param baud_rate: Baud rate (default: default configured ``baud_rate``).
        :param timeout: Communication timeout in seconds (default: default configured ``timeout``).
        """
        board, pin = self._get_board_and_pin(pin=pin,
                                             pin_type=PinType.ANALOG,
                                             board=board,
                                             board_type=board_type,
                                             baud_rate=baud_rate,
                                             timeout=timeout)
        board.analog[pin].write(value)

    @action
    def digital_write(self, pin: Union[int, str],
                      value: bool,
                      board: Optional[str] = None,
                      board_type: Optional[str] = None,
                      baud_rate: Optional[int] = None,
                      timeout: Optional[int] = None):
        """
        Write a value to a digital PIN.

        :param pin: PIN number or configured name.
        :param value: True (HIGH) or False (LOW).
        :param board: Board path or name (default: default configured ``board``).
        :param board_type: Board type. It can be 'mega', 'due' or 'nano' (default: configured ``board_type``).
        :param baud_rate: Baud rate (default: default configured ``baud_rate``).
        :param timeout: Communication timeout in seconds (default: default configured ``timeout``).
        """
        board, pin = self._get_board_and_pin(pin=pin,
                                             pin_type=PinType.DIGITAL,
                                             board=board,
                                             board_type=board_type,
                                             baud_rate=baud_rate,
                                             timeout=timeout)
        board.digital[pin].write(value)

    @action
    def pwm_write(self, pin: Union[int, str],
                  value: float,
                  board: Optional[str] = None,
                  board_type: Optional[str] = None,
                  baud_rate: Optional[int] = None,
                  timeout: Optional[int] = None):
        """
        Write a PWM value to a digital PIN.

        :param pin: PIN number or configured name.
        :param value: PWM real value normalized between 0 and 1.
        :param board: Board path or name (default: default configured ``board``).
        :param board_type: Board type. It can be 'mega', 'due' or 'nano' (default: configured ``board_type``).
        :param baud_rate: Baud rate (default: default configured ``baud_rate``).
        :param timeout: Communication timeout in seconds (default: default configured ``timeout``).
        """
        board, pin = self._get_board_and_pin(pin=pin,
                                             pin_type=PinType.DIGITAL,
                                             board=board,
                                             board_type=board_type,
                                             baud_rate=baud_rate,
                                             timeout=timeout)

        assert board.digital[pin].PWM_CAPABLE, 'PIN {} is not PWM capable'.format(pin)
        if board.digital[pin] != PWM:
            board.digital[pin].mode = PWM
            time.sleep(0.001)

        board.digital[pin].write(value)

    @action
    def get_measurement(self,
                        board: Optional[str] = None,
                        board_type: Optional[str] = None,
                        baud_rate: Optional[int] = None,
                        timeout: Optional[int] = None) -> Dict[str, float]:
        """
        Get a measurement from all the configured PINs.

        :param board: Board path or name (default: default configured ``board``)
        :param board_type: Board type. It can be 'mega', 'due' or 'nano' (default: configured ``board_type``).
        :param baud_rate: Baud rate (default: default configured ``baud_rate``)
        :param timeout: Communication timeout in seconds (default: default configured ``timeout``).
        :return: dict, where the keys are either the configured names of the PINs (see ``analog_pins`` configuration)
            or all the analog PINs (names will be in the format 'A0..A7' in that case), and the values will be the
            real values measured, either normalized between 0 and 1 if no conversion functions were provided, or
            transformed through the configured ``conv_functions``.
        """
        ret = {}
        board = self._get_board(board=board, board_type=board_type, baud_rate=baud_rate, timeout=timeout)

        for pin in board.analog:
            if self._pin_name_by_number[PinType.ANALOG] and \
                    pin.pin_number not in self._pin_name_by_number[PinType.ANALOG]:
                continue

            name = self._pin_name_by_number[PinType.ANALOG].get(pin.pin_number, 'A{}'.format(pin.pin_number))
            value = self._poll_value(pin=pin.pin_number, board=board, pin_type=PinType.ANALOG,
                                     timeout=timeout or self.timeout)

            if value is None:
                continue

            conv_function = self.conv_functions.get(name, self.conv_functions.get(pin.pin_number))
            if conv_function:
                value = conv_function(value)

            ret[name] = value

        return ret

    @action
    def close(self):
        for it in self._board_iterators.values():
            it.stop()

        for board in self._boards.values():
            board.exit()

        self._board_iterators = {}
        self._boards = {}


# vim:sw=4:ts=4:et:



import time

from typing import List, Tuple

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.ping import HostUpEvent, HostDownEvent
from platypush.utils.workers import Worker, Workers


class PingBackend(Backend):
    """
    This backend allows you to ping multiple remote hosts at regular intervals.

    Triggers:

        - :class:`platypush.message.event.ping.HostDownEvent` if a host stops responding ping requests
        - :class:`platypush.message.event.ping.HostUpEvent` if a host starts responding ping requests

    """

    class Pinger(Worker):
        def __init__(self, *args, **kwargs):
            self.timeout = kwargs.pop('timeout')
            self.count = kwargs.pop('count')
            super().__init__(*args, **kwargs)

        def process(self, host: str) -> Tuple[str, bool]:
            pinger = get_plugin('ping')
            response = pinger.ping(host, timeout=self.timeout, count=self.count).output
            return host, response['success'] is True

    def __init__(self, hosts: List[str], timeout: float = 5.0, interval: float = 60.0, count: int = 1,  *args, **kwargs):
        """
        :param hosts: List of IP addresses or host names to monitor.
        :param timeout: Ping timeout.
        :param interval: Interval between two scans.
        :param count: Number of pings per host. A host will be considered down
            if all the ping requests fail.
        """

        super().__init__(*args, **kwargs)
        self.hosts = {h: None for h in hosts}
        self.timeout = timeout
        self.interval = interval
        self.count = count

    def run(self):
        super().run()
        self.logger.info('Starting ping backend with {} hosts to monitor'.format(len(self.hosts)))

        while not self.should_stop():
            workers = Workers(10, self.Pinger, timeout=self.timeout, count=self.count)

            with workers:
                for host in self.hosts.keys():
                    workers.put(host)

            for response in workers.responses:
                host, is_up = response
                if is_up != self.hosts[host]:
                    if is_up:
                        self.bus.post(HostUpEvent(host))
                    else:
                        self.bus.post(HostDownEvent(host))

                self.hosts[host] = is_up

            time.sleep(self.interval)


# vim:sw=4:ts=4:et:



from platypush.plugins import Plugin, action


class LoggerPlugin(Plugin):
    """
    Plugin to log traces on the standard Platypush logger
    """

    @action
    def trace(self, msg, *args, **kwargs):
        """
        logger.trace wrapper
        """
        self.logger.trace(msg, *args, **kwargs)

    @action
    def debug(self, msg, *args, **kwargs):
        """
        logger.debug wrapper
        """
        self.logger.debug(msg, *args, **kwargs)

    @action
    def info(self, msg, *args, **kwargs):
        """
        logger.info wrapper
        """
        self.logger.info(msg, *args, **kwargs)

    @action
    def warning(self, msg, *args, **kwargs):
        """
        logger.warning wrapper
        """
        self.logger.warning(msg, *args, **kwargs)

    @action
    def error(self, msg, *args, **kwargs):
        """
        logger.error wrapper
        """
        self.logger.error(msg, *args, **kwargs)


# vim:sw=4:ts=4:et:




import json
import logging
import os
import re

from platypush.config import Config
from platypush.backend.http.app import template_folder


class HttpUtils(object):
    log = logging.getLogger(__name__)

    @staticmethod
    def widget_columns_to_html_class(columns):
        if not isinstance(columns, int):
            try:
                columns = int(columns)
            except ValueError:
                raise RuntimeError('columns should be a number, got {} ({})'.format(type(columns), columns))

        if 1 <= columns <= 12:
            return 'col-{}'.format(columns)

        raise RuntimeError('Constraint violation: should be 1 <= columns <= 12, ' +
                           'got columns={}'.format(columns))

    @staticmethod
    def search_directory(directory, *extensions, recursive=False):
        files = []

        if recursive:
            for root, subdirs, files in os.walk(directory):
                for file in files:
                    if not extensions or os.path.splitext(file)[1].lower() in extensions:
                        files.append(os.path.join(root, file))
        else:
            for file in os.listdir(directory):
                if not extensions or os.path.splitext(file)[1].lower() in extensions:
                    files.append(os.path.join(directory, file))

        return files

    @classmethod
    def search_web_directory(cls, directory, *extensions):
        directory = os.path.abspath(os.path.expanduser(directory))
        resource_dirs = (Config.get('backend.http') or {}).get('resource_dirs', {})
        resource_path = None
        uri = ''

        for name, resource_path in resource_dirs.items():
            resource_path = os.path.abspath(os.path.expanduser(resource_path))
            if directory.startswith(resource_path):
                subdir = re.sub('^{}(.*)$'.format(resource_path),
                                '\\1', directory)
                uri = '/resources/' + name
                break

        if not uri:
            raise RuntimeError(('Directory {} not found among the available ' +
                               'static resources on the webserver').format(
                                   directory))

        results = [
            re.sub('^{}(.*)$'.format(resource_path), uri + '\\1', path)
            for path in cls.search_directory(directory, *extensions)
        ]

        return results

    @classmethod
    def to_json(cls, data):
        def json_parse(x):
            if type(x) == __import__('datetime').timedelta:
                return x.days * 24 * 60 * 60 + x.seconds + x.microseconds / 1e6

            # Ignore non-serializable attributes
            cls.log.warning('Non-serializable attribute type "{}": {}'.format(type(x), x))
            return None

        if isinstance(data, type({}.keys())):
            # Convert dict_keys to list before serializing
            data = list(data)
        return json.dumps(data, default=json_parse)

    @classmethod
    def from_json(cls, data):
        return json.loads(data)

    @classmethod
    def get_config(cls, attr):
        return Config.get(attr)

    @classmethod
    def plugin_name_to_tag(cls, module_name):
        return module_name.replace('.','-')

    @classmethod
    def find_templates_in_dir(cls, directory):
        return [
            os.path.join(directory, file)
            for root, path, files in os.walk(os.path.abspath(os.path.join(template_folder, directory)))
            for file in files
            if file.endswith('.html') or file.endswith('.htm')
        ]

    @classmethod
    def readfile(cls, file):
        with open(file) as f:
            return f.read()

    @classmethod
    def isfile(cls, *path):
        path = path[0] if len(path) == 1 else os.path.join(*path)
        return os.path.isfile(path)

# vim:sw=4:ts=4:et:



import datetime
import dateutil.parser

from platypush.plugins import action
from platypush.plugins.http.request import HttpRequestPlugin

class HttpRequestOtaBookingPlugin(HttpRequestPlugin):
    """ Plugin to send requests to the Booking Hub API """

    def __init__(self, hotel_id, token, timeout=5, **kwargs):
        self.hotel_id = hotel_id
        self.token = token
        self.timeout = timeout


    @action
    def get_reservations(self, day='today'):
        url = 'https://hub-api.booking.com/v1/hotels/{}/reservations' \
            .format(self.hotel_id)

        today = datetime.date.today().isoformat()
        if day == 'today': day = today

        headers = { 'X-Booking-Auth-Token': self.token }
        params = { 'checkin': day }

        response = self.get(url, headers=headers, params=params,
                            output='json', timeout=self.timeout)

        reservations = [res for res in response.output
                        if res['status'] != 'CANCELLED']

        response.output = {
            'reservations': reservations,
            'n_reservations': len(reservations),
        }

        return response


# vim:sw=4:ts=4:et:




from typing import Optional

from platypush.message.event import Event


class LinodeEvent(Event):
    pass


class LinodeInstanceStatusChanged(LinodeEvent):
    """
    Event triggered when the status of a Linode instance changes.
    """
    def __init__(self, instance: str, status: str, old_status: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, instance=instance, status=status, old_status=old_status, **kwargs)


# vim:sw=4:ts=4:et:



import os

from flask import Blueprint, render_template

from platypush.backend.http.app import template_folder, static_folder
from platypush.backend.http.app.utils import authenticate, get_websocket_port

from platypush.backend.http.utils import HttpUtils
from platypush.config import Config

panel = Blueprint('plugin', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    panel,
]


@panel.route('/plugin/<plugin>', methods=['GET'])
@authenticate()
def plugin_route(plugin):
    """ Route to the plugin pane template """
    js_folder = os.path.abspath(os.path.join(template_folder, '..', 'static', 'js'))
    style_folder = os.path.abspath(os.path.join(template_folder, '..', 'static', 'css', 'dist'))
    template_file = os.path.join(template_folder, 'plugins', plugin, 'index.html')
    script_file = os.path.join(js_folder, 'plugins', plugin, 'index.js')
    style_file = os.path.join(style_folder, 'webpanel', 'plugins', plugin+'.css')

    conf = Config.get(plugin) or {}
    if os.path.isfile(template_file):
        conf['_template_file'] = '/' + '/'.join(template_file.split(os.sep)[-3:])

    if os.path.isfile(script_file):
        conf['_script_file'] = '/'.join(script_file.split(os.sep)[-4:])

    if os.path.isfile(style_file):
        conf['_style_file'] = 'css/dist/' + style_file[len(style_folder)+1:]

    http_conf = Config.get('backend.http')
    return render_template('plugin.html',
                           plugin=plugin,
                           conf=conf,
                           template=conf.get('_template_file', {}),
                           script=conf.get('_script_file', {}),
                           style=conf.get('_style_file', {}),
                           utils=HttpUtils,
                           token=Config.get('token'),
                           websocket_port=get_websocket_port(),
                           template_folder=template_folder,
                           static_folder=static_folder,
                           has_ssl=http_conf.get('ssl_cert') is not None)


# vim:sw=4:ts=4:et:



from platypush.message.event.http import HttpEvent

class NewReservationEvent(HttpEvent):
    def __init__(self, request, response, *args, **kwargs):
        super().__init__(request=request, response=response, *args, **kwargs)


# vim:sw=4:ts=4:et:




import threading

from typing import Optional, List, Any, Dict

from platypush.plugins.mqtt import MqttPlugin, action


class ZigbeeMqttPlugin(MqttPlugin):
    """
    This plugin allows you to interact with Zigbee devices over MQTT through any Zigbee sniffer and
    `zigbee2mqtt <https://www.zigbee2mqtt.io/>`_.

    In order to get started you'll need:

        - A Zigbee USB adapter/sniffer (in this example I'll use the `CC2531 <https://hackaday.io/project/163487-zigbee-cc2531-smart-home-usb-adapter>`_.
        - A Zigbee debugger/emulator + downloader cable (only to flash the firmware).

    Instructions:

        - Install `cc-tool <https://github.com/dashesy/cc-tool>`_ either from sources or from a package manager.
        - Connect the Zigbee to your PC/RaspberryPi in this way: USB -> CC debugger -> downloader cable -> CC2531 -> USB.
          The debugger and the adapter should be connected _at the same time_. If the later ``cc-tool`` command throws up
          an error, put the device in sync while connected by pressing the _Reset_ button on the debugger.
        - Check where the device is mapped. On Linux it will usually be ``/dev/ttyACM0``.
        - Download the latest `Z-Stack firmware <https://github.com/Koenkk/Z-Stack-firmware/tree/master/coordinator>`_ to your device.
          Instructions for a CC2531 device:

          .. code-block:: shell

              wget https://github.com/Koenkk/Z-Stack-firmware/raw/master/coordinator/Z-Stack_Home_1.2/bin/default/CC2531_DEFAULT_20190608.zip
              unzip CC2531_DEFAULT_20190608.zip
              [sudo] cc-tool -e -w CC2531ZNP-Prod.hex

        - You can disconnect your debugger and downloader cable once the firmware is flashed.

        - Install ``zigbee2mqtt``. First install a node/npm environment, then either install ``zigbee2mqtt`` manually or through your
          package manager. Manual instructions:

          .. code-block:: shell

              # Clone zigbee2mqtt repository
              [sudo] git clone https://github.com/Koenkk/zigbee2mqtt.git /opt/zigbee2mqtt
              [sudo] chown -R pi:pi /opt/zigbee2mqtt  # Or whichever is your user

              # Install dependencies (as user "pi")
              cd /opt/zigbee2mqtt
              npm install

        - You need to have an MQTT broker running somewhere. If not, you can install
            `Mosquitto <https://mosquitto.org/>`_ through your package manager on any device in your network.

        - Edit the ``/opt/zigbee2mqtt/data/configuration.yaml`` file to match the configuration of your MQTT broker:

          .. code-block:: yaml

              # MQTT settings
              mqtt:
                  # MQTT base topic for zigbee2mqtt MQTT messages
                  base_topic: zigbee2mqtt
                  # MQTT server URL
                  server: 'mqtt://localhost'
                  # MQTT server authentication, uncomment if required:
                  # user: my_user
                  # password: my_password

        - Also make sure that ``permit_join`` is set to ``True``, in order to allow Zigbee devices to join the network
          while you're configuring it. It's equally important to set ``permit_join`` to ``False`` once you have
          configured your network, to prevent accidental/malignant joins from outer Zigbee devices.

        - Start the ``zigbee2mqtt`` daemon on your device (the
          `official documentation <https://www.zigbee2mqtt.io/getting_started/running_zigbee2mqtt.html#5-optional-running-as-a-daemon-with-systemctl>`_
          also contains instructions on how to configure it as a ``systemd`` service:

          .. code-block:: shell

              cd /opt/zigbee2mqtt
              npm start

        - If you have Zigbee devices that are paired to other bridges, unlink them or do a factory reset to pair them
          to your new bridge.

        - If it all goes fine, once the daemon is running and a new device is found you should see traces like this in
          the output of ``zigbee2mqtt``::

            zigbee2mqtt:info  2019-11-09T12:19:56: Successfully interviewed '0x00158d0001dc126a', device has successfully been paired

        - You are now ready to use this integration.

    Requires:

        * **paho-mqtt** (``pip install paho-mqtt``)

    """

    def __init__(self, host: str = 'localhost', port: int = 1883, base_topic: str = 'zigbee2mqtt', timeout: int = 60,
                 tls_certfile: Optional[str] = None, tls_keyfile: Optional[str] = None,
                 tls_version: Optional[str] = None, tls_ciphers: Optional[str] = None,
                 username: Optional[str] = None, password: Optional[str] = None, **kwargs):
        """
        :param host: Default MQTT broker where ``zigbee2mqtt`` publishes its messages (default: ``localhost``).
        :param port: Broker listen port (default: 1883).
        :param base_topic: Topic prefix, as specified in ``/opt/zigbee2mqtt/data/configuration.yaml``
            (default: '``base_topic``').
        :param timeout: If the command expects from a response, then this timeout value will be used
            (default: 60 seconds).
        :param tls_cafile: If the connection requires TLS/SSL, specify the certificate authority file
            (default: None)
        :param tls_certfile: If the connection requires TLS/SSL, specify the certificate file (default: None)
        :param tls_keyfile: If the connection requires TLS/SSL, specify the key file (default: None)
        :param tls_version: If the connection requires TLS/SSL, specify the minimum TLS supported version
            (default: None)
        :param tls_ciphers: If the connection requires TLS/SSL, specify the supported ciphers (default: None)
        :param username: If the connection requires user authentication, specify the username (default: None)
        :param password: If the connection requires user authentication, specify the password (default: None)
        """

        super().__init__(host=host, port=port, tls_certfile=tls_certfile, tls_keyfile=tls_keyfile,
                         tls_version=tls_version, tls_ciphers=tls_ciphers, username=username,
                         password=password, **kwargs)

        self.base_topic = base_topic
        self.timeout = timeout

    def _mqtt_args(self, host: Optional[str] = None, **kwargs):
        if not host:
            return {
                'host': self.host,
                'port': self.port,
                'timeout': self.timeout,
                'tls_certfile': self.tls_certfile,
                'tls_keyfile': self.tls_keyfile,
                'tls_version': self.tls_version,
                'tls_ciphers': self.tls_ciphers,
                'username': self.username,
                'password': self.password,
            }

        return kwargs

    def _topic(self, topic):
        return self.base_topic + '/' + topic

    @action
    def devices(self, **kwargs) -> List[Dict[str, Any]]:
        """
        Get the list of devices registered to the service.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).

        :return: List of paired devices. Example output:

        .. code-block:: json

            [
                {
                    "dateCode": "20190608",
                    "friendly_name": "Coordinator",
                    "ieeeAddr": "0x00123456789abcde",
                    "lastSeen": 1579640601215,
                    "networkAddress": 0,
                    "softwareBuildID": "zStack12",
                    "type": "Coordinator"
                },
                {
                    "dateCode": "20160906",
                    "friendly_name": "My Lightbulb",
                    "hardwareVersion": 1,
                    "ieeeAddr": "0x00123456789abcdf",
                    "lastSeen": 1579595191623,
                    "manufacturerID": 4107,
                    "manufacturerName": "Philips",
                    "model": "8718696598283",
                    "modelID": "LTW013",
                    "networkAddress": 52715,
                    "powerSource": "Mains (single phase)",
                    "softwareBuildID": "1.15.2_r19181",
                    "type": "Router"
                }
            ]

        """
        return self.publish(
            topic=self._topic('bridge/config/devices/get'), msg='',
            reply_topic=self._topic('bridge/config/devices'),
            **self._mqtt_args(**kwargs))

    def _permit_join_timeout_callback(self, permit: bool, **kwargs):
        def callback():
            self.logger.info('Restoring permit_join state to {}'.format(permit))
            self.permit_join(permit, **kwargs)
        return callback

    @action
    def permit_join(self, permit: bool = True, timeout: Optional[float] = None, **kwargs):
        """
        Enable/disable devices from joining the network. This is not persistent (will not be saved to
        ``configuration.yaml``).

        :param permit: Set to True to allow joins, False otherwise.
        :param timeout: Allow/disallow joins only for this amount of time.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.publish(topic=self._topic('bridge/config/permit_join'), msg=permit, **self._mqtt_args(**kwargs))
        if timeout:
            threading.Timer(timeout, self._permit_join_timeout_callback(not permit, **kwargs)).start()

    @action
    def reset(self, **kwargs):
        """
        Reset the adapter.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.publish(topic=self._topic('bridge/config/reset'), msg='', **self._mqtt_args(**kwargs))

    @action
    def factory_reset(self, **kwargs):
        """
        Perform a factory reset of the device. Of course, you should only do it if you know what you're doing,
        as you will lose all the paired devices and may also lose the Z-Stack firmware.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.publish(topic=self._topic('bridge/config/factory_reset'), msg='', **self._mqtt_args(**kwargs))

    @action
    def log_level(self, level: str, **kwargs):
        """
        Change the log level at runtime. This change will not be persistent.

        :param level: Possible values: 'debug', 'info', 'warn', 'error'.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.publish(topic=self._topic('bridge/config/log_level'), msg=level, **self._mqtt_args(**kwargs))

    @action
    def device_set_option(self, device: str, option: str, value: Any, **kwargs):
        """
        Change the options of a device. Options can only be changed, not added or deleted.

        :param device: Display name of the device.
        :param option: Option name.
        :param value: New value.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.publish(topic=self._topic('bridge/config/device_options'), msg={
            'friendly_name': device,
            'options': {
                option: value,
            }
        }, **self._mqtt_args(**kwargs))

    @action
    def device_remove(self, device: str, force: bool = False, **kwargs):
        """
        Remove a device from the network.

        :param device: Display name of the device.
        :param force: Force the remove also if the removal wasn't acknowledged by the device. Note: a forced remove
            only removes the entry from the internal database, but the device is likely to connect again when
            restarted unless it's factory reset (default: False).
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        topic = self._topic('bridge/config/{}remove'.format('force_' if force else ''))
        self.publish(topic=topic, msg=device, **self._mqtt_args(**kwargs))

    @action
    def device_ban(self, device: str, **kwargs):
        """
        Ban a device from the network.

        :param device: Display name of the device.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.publish(topic=self._topic('bridge/config/ban'), msg=device, **self._mqtt_args(**kwargs))

    @action
    def device_whitelist(self, device: str, **kwargs):
        """
        Whitelist a device on the network. Note: once at least a device is whitelisted, all the other non-whitelisted
        devices will be removed from the network.

        :param device: Display name of the device.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.publish(topic=self._topic('bridge/config/whitelist'), msg=device, **self._mqtt_args(**kwargs))

    @action
    def device_rename(self, name: str, device: Optional[str] = None, **kwargs):
        """
        Rename a device on the network.

        :param name: New name.
        :param device: Current name of the device to rename. If no name is specified then the rename will
            affect the last device that joined the network.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        if name == device:
            self.logger.info('Old and new name are the same: nothing to do')
            return

        # noinspection PyUnresolvedReferences
        devices = self.devices().output
        assert not [dev for dev in devices if dev.get('friendly_name') == name], \
            'A device named {} already exists on the network'.format(name)

        self.publish(
            topic=self._topic('bridge/config/rename{}'.format('_last' if not device else '')),
            msg={'old': device, 'new': name} if device else name,
            **self._mqtt_args(**kwargs))

    # noinspection PyShadowingBuiltins
    @action
    def device_get(self, device: str, property: Optional[str] = None, **kwargs) -> Dict[str, Any]:
        """
        Get the properties of a device. The returned keys vary depending on the device. For example, a light bulb
        may have the "``state``" and "``brightness``" properties, while an environment sensor may have the
        "``temperature``" and "``humidity``" properties, and so on.

        :param device: Display name of the device.
        :param property: Name of the property that should be retrieved (default: all).
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        :return: Key->value map of the device properties.
        """
        properties = self.publish(topic=self._topic(device + '/get'),
                                  reply_topic=self._topic(device), msg='', **self._mqtt_args(**kwargs)).output

        if property:
            assert property in properties, 'No such property: ' + property
            return {property: properties[property]}

        return properties

    # noinspection PyShadowingBuiltins,DuplicatedCode
    @action
    def device_set(self, device: str, property: str, value: Any, **kwargs):
        """
        Set a properties on a device. The compatible properties vary depending on the device. For example, a light bulb
        may have the "``state``" and "``brightness``" properties, while an environment sensor may have the
        "``temperature``" and "``humidity``" properties, and so on.

        :param device: Display name of the device.
        :param property: Name of the property that should be set.
        :param value: New value of the property.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        properties = self.publish(topic=self._topic(device + '/set'),
                                  reply_topic=self._topic(device),
                                  msg={property: value}, **self._mqtt_args(**kwargs)).output

        if property:
            assert property in properties, 'No such property: ' + property
            return {property: properties[property]}

        return properties

    @action
    def device_groups(self, device: str, **kwargs) -> List[int]:
        """
        List the groups a given device belongs to.

        :param device: Display name of the device.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        :return: List of group IDs the device is linked to.
        """

        return self.publish(topic=self._topic('bridge/device/{}/get_group_membership'.format(device)),
                            reply_topic=self._topic(device), msg=device, **self._mqtt_args(**kwargs)).\
            output.get('group_list', [])

    @action
    def groups(self, **kwargs):
        """
        Get the groups registered on the device.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        groups = self.publish(topic=self._topic('bridge/config/groups'), msg={},
                              reply_topic=self._topic('bridge/log'),
                              **self._mqtt_args(**kwargs)).output.get('message', [])

        # noinspection PyUnresolvedReferences
        devices = {
            device['ieeeAddr']: device
            for device in self.devices(**kwargs).output
        }

        return [
            {
                'id': group['ID'],
                'friendly_name': group['friendly_name'],
                'optimistic': group.get('optimistic', False),
                'devices': [
                    devices[device.split('/')[0]]
                    for device in group.get('devices', [])
                ]
            }
            for group in groups
        ]

    @action
    def group_add(self, name: str, id: Optional[int] = None, **kwargs):
        """
        Add a new group.

        :param name: Display name of the group.
        :param id: Optional numeric ID (default: auto-generated).
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        args = {'friendly_name': name}
        if id is not None:
            args['id'] = id

        self.publish(topic=self._topic('bridge/config/add_group'), msg=args, **self._mqtt_args(**kwargs))

    # noinspection PyShadowingBuiltins,DuplicatedCode
    @action
    def group_set(self, group: str, property: str, value: Any, **kwargs):
        """
        Set a properties on a group. The compatible properties vary depending on the devices on the group.
        For example, a light bulb may have the "``state``" (with values ``"ON"`` and ``"OFF"``) and "``brightness``"
        properties, while an environment sensor may have the "``temperature``" and "``humidity``" properties, and so on.

        :param group: Display name of the group.
        :param property: Name of the property that should be set.
        :param value: New value of the property.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        properties = self.publish(topic=self._topic(group + '/set'),
                                  reply_topic=self._topic(group),
                                  msg={property: value}, **self._mqtt_args(**kwargs)).output

        if property:
            assert property in properties, 'No such property: ' + property
            return {property: properties[property]}

        return properties

    @action
    def group_rename(self, name: str, group: str, **kwargs):
        """
        Rename a group.

        :param name: New name.
        :param group: Current name of the group to rename.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        if name == group:
            self.logger.info('Old and new name are the same: nothing to do')
            return

        # noinspection PyUnresolvedReferences
        groups = {group.get('friendly_name'): group for group in self.groups().output}
        assert name not in groups, 'A group named {} already exists on the network'.format(name)

        self.publish(
            topic=self._topic('bridge/config/rename'),
            msg={'old': group, 'new': name} if group else name,
            **self._mqtt_args(**kwargs))

    @action
    def group_remove(self, name: str, **kwargs):
        """
        Remove a group.

        :param name: Display name of the group.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.publish(topic=self._topic('bridge/config/remove_group'), msg=name,
                     **self._mqtt_args(**kwargs))

    @action
    def group_add_device(self, group: str, device: str, **kwargs):
        """
        Add a device to a group.

        :param group: Display name of the group.
        :param device: Display name of the device to be added.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.publish(topic=self._topic('bridge/group/{}/add'.format(group)),
                     msg=device, **self._mqtt_args(**kwargs))

    @action
    def group_remove_device(self, group: str, device: Optional[str] = None, **kwargs):
        """
        Remove a device from a group.

        :param group: Display name of the group.
        :param device: Display name of the device to be removed. If none is specified then all the devices registered
            to the specified group will be removed.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.publish(topic=self._topic('bridge/group/{}/remove{}'.format(group, '_all' if not device else '')),
                     msg=device, **self._mqtt_args(**kwargs))

    @action
    def bind_devices(self, source: str, target: str, endpoint: Optional[str] = None, **kwargs):
        """
        Bind two devices. Binding makes it possible that devices can directly control each other without the
        intervention of zigbee2mqtt or any home automation software. You may want to use this feature to bind
        for example an IKEA/Philips Hue dimmer switch to a light bulb, or a Zigbee remote to a thermostat.
        Read more on the `zigbee2mqtt binding page <https://www.zigbee2mqtt.io/information/binding.html>`_.

        :param source: Name of the source device. It can also be a group name, although the support is
            `still experimental <https://www.zigbee2mqtt.io/information/binding.html#binding-a-group>`_.
        :param target: Name of the target device.
        :param endpoint: The target may support multiple endpoints (e.g. 'left', 'down', 'up' etc.). If so,
            you can bind the source to a specific endpoint on the target device.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.publish(topic=self._topic('bridge/bind/' + source + ('/' + endpoint if endpoint else '')),
                     msg=target, **self._mqtt_args(**kwargs))

    @action
    def unbind_devices(self, source: str, target: str, **kwargs):
        """
        Un-bind two devices.

        :param source: Name of the source device.
        :param target: Name of the target device.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.publish(topic=self._topic('bridge/unbind/' + source),
                     msg=target, **self._mqtt_args(**kwargs))


# vim:sw=4:ts=4:et:



from typing import Optional

from platypush.message.event import Event


class AlarmEvent(Event):
    def __init__(self, name: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, name=name, **kwargs)


class AlarmStartedEvent(AlarmEvent):
    """
    Triggered when an alarm starts.
    """
    pass


class AlarmEndedEvent(AlarmEvent):
    """
    Triggered when an alarm stops.
    """
    pass


class AlarmDismissedEvent(AlarmEndedEvent):
    """
    Triggered when an alarm is dismissed.
    """
    pass


class AlarmSnoozedEvent(AlarmEvent):
    """
    Triggered when an alarm is snoozed.
    """
    pass


class AlarmTimeoutEvent(AlarmEndedEvent):
    """
    Triggered when an alarm times out.
    """
    pass


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class LightOnEvent(Event):
    """
    Event triggered when a light on event is detected
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class LightOffEvent(Event):
    """
    Event triggered when a light off event is detected
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# vim:sw=4:ts=4:et:




from platypush.plugins import action
from platypush.plugins.gpio.sensor import GpioSensorPlugin


class GpioSensorAccelerometerPlugin(GpioSensorPlugin):
    """
    Plugin to interact with an accelerometer sensor and get X,Y,Z position.
    Tested with Adafruit LIS3DH accelerometer (https://www.adafruit.com/product/2809)
    with Raspberry Pi over I2C connection.

    Requires:

        * ``Adafruit_Python_GPIO`` (``pip install Adafruit_Python_GPIO``)
    """

    def __init__(self, g=4, precision=None, *args, **kwargs):
        """
        Only LIS3DH in I2C mode is currently supported: https://learn.adafruit.com/assets/59080.

        :param g: Accelerometer range as a multiple of G - can be 2G, 4G, 8G or 16G
        :type g: int

        :param precision: If set, the position values will be rounded to the specified number of decimal digits
            (default: no rounding)
        :type precision: int
        """

        super().__init__(*args, **kwargs)
        from platypush.plugins.gpio.sensor.accelerometer.lib.LIS3DH import LIS3DH

        if g == 2:
            self.g = LIS3DH.RANGE_2G
        elif g == 4:
            self.g = LIS3DH.RANGE_4G
        elif g == 8:
            self.g = LIS3DH.RANGE_8G
        elif g == 16:
            self.g = LIS3DH.RANGE_16G
        else:
            raise RuntimeError('Invalid G range: {}'.format(g))

        self.precision = precision
        self.sensor = LIS3DH()
        self.sensor.setRange(self.g)

    @action
    def get_measurement(self):
        """
        Extends :func:`.GpioSensorPlugin.get_measurement`

        :returns: The sensor's current position as a dictionary with the three components (x,y,z) in degrees, each
            between -90 and 90
        """

        values = [
            (pos*100 if self.precision is None else round(pos*100, self.precision))
            for pos in (self.sensor.getX(), self.sensor.getY(), self.sensor.getZ())
        ]

        return {
            'name': 'position',
            'value': {
                'x': values[0], 'y': values[1], 'z': values[2]
            }
        }


# vim:sw=4:ts=4:et:



import json
import socket
import threading

from platypush.config import Config
from platypush.context import get_backend
from platypush.plugins import Plugin, action


class MusicSnapcastPlugin(Plugin):
    """
    Plugin to interact with a [Snapcast](https://github.com/badaix/snapcast)
    instance, control clients mute status, volume, playback etc.

    See https://github.com/badaix/snapcast/blob/master/doc/json_rpc_api/v2_0_0.md
    for further reference about the returned API types.
    """

    _DEFAULT_SNAPCAST_PORT = 1705
    _SOCKET_EOL = '\r\n'.encode()

    def __init__(self, host='localhost', port=_DEFAULT_SNAPCAST_PORT,
                 *args, **kwargs):
        """
        :param host: Default Snapcast server host (default: localhost)
        :type host: str

        :param port: Default Snapcast server control port (default: 1705)
        :type port: int
        """

        super().__init__(**kwargs)

        self.host = host
        self.port = port
        self._latest_req_id = 0
        self._latest_req_id_lock = threading.RLock()

        backend = get_backend('music.snapcast')
        backend_hosts = backend.hosts if backend else [self.host]
        backend_ports = backend.ports if backend else [self.port]
        self.backend_hosts = backend_hosts
        self.backend_ports = backend_ports

    def _get_req_id(self):
        with self._latest_req_id_lock:
            self._latest_req_id += 1
            return self._latest_req_id

    def _connect(self, host=None, port=None):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.logger.info('Connecting to Snapcast host {}:{}'.format(host, port))
        sock.connect((host or self.host, port or self.port))
        return sock

    @classmethod
    def _send(cls, sock, req):
        if isinstance(req, dict):
            req = json.dumps(req)
        if isinstance(req, str):
            req = req.encode()
        if not isinstance(req, bytes):
            raise RuntimeError('Unsupported type {} for Snapcast request: {}'.
                               format(type(req), req))

        sock.send(req + cls._SOCKET_EOL)

    @classmethod
    def _recv(cls, sock):
        buf = b''
        while buf[-2:] != cls._SOCKET_EOL:
            buf += sock.recv(1)
        return json.loads(buf.decode().strip()).get('result')

    def _get_group(self, sock, group):
        for g in self._status(sock).get('groups', []):
            if group == g.get('id') or group == g.get('name'):
                return g

    def _get_client(self, sock, client):
        for g in self._status(sock).get('groups', []):
            clients = g.get('clients', [])

            for c in clients:
                if client == c.get('id') or \
                        client == c.get('name') or \
                        client == c.get('host', {}).get('name') or \
                        client == c.get('host', {}).get('ip'):
                    c['group_id'] = g.get('id')
                    return c

    def _status(self, sock):
        request = {
            'id': self._get_req_id(),
            'jsonrpc':'2.0',
            'method':'Server.GetStatus'
        }

        self._send(sock, request)
        return (self._recv(sock) or {}).get('server', {})

    @action
    def status(self, host=None, port=None, client=None, group=None):
        """
        Get the status either of a Snapcast server, client or group

        :param host: Snapcast server to query (default: default configured host)
        :type host: str

        :param port: Snapcast server port (default: default configured port)
        :type port: int

        :param client: Client ID or name (default: None)
        :type client: str

        :param group: Group ID or name (default: None)
        :type group: str

        :returns: dict.

        Example::

            "output": {
                "groups": [
                    {
                    "clients": [
                        {
                        "config": {
                            "instance": 1,
                            "latency": 0,
                            "name": "",
                            "volume": {
                            "muted": false,
                            "percent": 96
                            }
                        },
                        "connected": true,
                        "host": {
                            "arch": "x86_64",
                            "ip": "YOUR_IP",
                            "mac": "YOUR_MAC",
                            "name": "YOUR_NAME",
                            "os": "YOUR_OS"
                        },
                        "id": "YOUR_ID",
                        "lastSeen": {
                            "sec": 1546648311,
                            "usec": 86011
                        },
                        "snapclient": {
                            "name": "Snapclient",
                            "protocolVersion": 2,
                            "version": "0.15.0"
                        }
                        }
                    ],
                    "id": "YOUR_ID",
                    "muted": false,
                    "name": "",
                    "stream_id": "mopidy"
                    }
                ],
                "server": {
                    "host": {
                    "arch": "armv7l",
                    "ip": "",
                    "mac": "",
                    "name": "YOUR_NAME",
                    "os": "YOUR_OS"
                    },
                    "snapserver": {
                    "controlProtocolVersion": 1,
                    "name": "Snapserver",
                    "protocolVersion": 1,
                    "version": "0.15.0"
                    }
                },
                "streams": [
                    {
                        "id": "mopidy",
                        "meta": {
                            "STREAM": "mopidy"
                        },
                        "status": "playing",
                        "uri": {
                            "fragment": "",
                            "host": "",
                            "path": "/tmp/snapfifo",
                            "query": {
                            "buffer_ms": "20",
                            "codec": "pcm",
                            "name": "mopidy",
                            "sampleformat": "48000:16:2"
                            },
                            "raw": "pipe:////tmp/snapfifo?buffer_ms=20&codec=pcm&name=mopidy&sampleformat=48000:16:2",
                            "scheme": "pipe"
                        }
                    }
                ]
            }

        """

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            if client:
                return self._get_client(sock, client)
            if group:
                return self._get_group(sock, group)

            return self._status(sock)
        finally:
            try: sock.close()
            except: pass


    @action
    def mute(self, client=None, group=None, mute=None, host=None, port=None):
        """
        Set the mute status of a connected client or group

        :param client: Client name or ID to mute
        :type client: str

        :param group: Group ID to mute
        :type group: str

        :param mute: Mute status. If not set, the mute status of the selected
            client/group will be toggled.
        :type mute: bool

        :param host: Snapcast server to query (default: default configured host)
        :type host: str

        :param port: Snapcast server port (default: default configured port)
        :type port: int
        """

        if not client and not group:
            raise RuntimeError('Please specify either a client or a group')

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            request = {
                'id': self._get_req_id(),
                'jsonrpc':'2.0',
                'method': 'Group.SetMute' if group else 'Client.SetVolume',
                'params': {}
            }

            if group:
                group = self._get_group(sock, group)
                cur_muted = group['muted']
                request['params']['id'] = group['id']
                request['params']['mute'] = not cur_muted if mute is None else mute
            else:
                client = self._get_client(sock, client)
                cur_muted = client['config']['volume']['muted']
                request['params']['id'] = client['id']
                request['params']['volume'] = {}
                request['params']['volume']['percent'] = client['config']['volume']['percent']
                request['params']['volume']['muted'] = not cur_muted if mute is None else mute

            self._send(sock, request)
            return self._recv(sock)
        finally:
            try: sock.close()
            except: pass

    @action
    def volume(self, client, volume=None, delta=None, mute=None, host=None,
               port=None):
        """
        Set the volume of a connected client

        :param client: Client name or ID
        :type client: str

        :param volume: Absolute volume to set between 0 and 100
        :type volume: int

        :param delta: Relative volume change in percentage (e.g. +10 or -10)
        :type delta: int

        :param mute: Set to true or false if you want to toggle the muted status
        :type mute: bool

        :param host: Snapcast server (default: default configured host)
        :type host: str

        :param port: Snapcast server port (default: default configured port)
        :type port: int
        """

        if volume is None and delta is None and mute is None:
            raise RuntimeError('Please specify either an absolute volume or ' +
                               'relative delta')

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            request = {
                'id': self._get_req_id(),
                'jsonrpc':'2.0',
                'method': 'Client.SetVolume',
                'params': {}
            }

            client = self._get_client(sock, client)
            cur_volume = int(client['config']['volume']['percent'])
            cur_mute = bool(client['config']['volume']['muted'])

            if volume is not None:
                volume = int(volume)
            elif delta is not None:
                volume = cur_volume + int(delta)

            if volume is not None:
                if volume > 100: volume = 100
                if volume < 0: volume = 0
            else:
                volume = cur_volume

            if mute is None:
                mute = cur_mute

            request['params']['id'] = client['id']
            request['params']['volume'] = {}
            request['params']['volume']['percent'] = volume
            request['params']['volume']['muted'] = mute
            self._send(sock, request)
            return self._recv(sock)
        finally:
            try: sock.close()
            except: pass

    @action
    def set_client_name(self, client, name, host=None, port=None):
        """
        Set/change the name of a connected client

        :param client: Current client name or ID to rename
        :type client: str

        :param name: New name
        :type name: str

        :param host: Snapcast server (default: default configured host)
        :type host: str

        :param port: Snapcast server port (default: default configured port)
        :type port: int
        """

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            request = {
                'id': self._get_req_id(),
                'jsonrpc':'2.0',
                'method': 'Client.SetName',
                'params': {}
            }

            client = self._get_client(sock, client)
            request['params']['id'] = client['id']
            request['params']['name'] = name
            self._send(sock, request)
            return self._recv(sock)
        finally:
            try: sock.close()
            except: pass

    @action
    def set_latency(self, client, latency, host=None, port=None):
        """
        Set/change the latency of a connected client

        :param client: Client name or ID
        :type client: str

        :param latency: New latency in milliseconds
        :type latency: float

        :param host: Snapcast server (default: default configured host)
        :type host: str

        :param port: Snapcast server port (default: default configured port)
        :type port: int
        """

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            request = {
                'id': self._get_req_id(),
                'jsonrpc':'2.0',
                'method': 'Client.SetLatency',
                'params': {
                    'latency': latency
                }
            }

            client = self._get_client(sock, client)
            request['params']['id'] = client['id']
            self._send(sock, request)
            return self._recv(sock)
        finally:
            try: sock.close()
            except: pass

    @action
    def delete_client(self, client, host=None, port=None):
        """
        Delete a client from the Snapcast server

        :param client: Client name or ID
        :type client: str

        :param host: Snapcast server (default: default configured host)
        :type host: str

        :param port: Snapcast server port (default: default configured port)
        :type port: int
        """

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            request = {
                'id': self._get_req_id(),
                'jsonrpc':'2.0',
                'method': 'Server.DeleteClient',
                'params': {}
            }

            client = self._get_client(sock, client)
            request['params']['id'] = client['id']
            self._send(sock, request)
            return self._recv(sock)
        finally:
            try: sock.close()
            except: pass

    @action
    def group_set_clients(self, group, clients, host=None, port=None):
        """
        Sets the clients for a group on a Snapcast server

        :param group: Group name or ID
        :type group: str

        :param clients: List of client names or IDs
        :type clients: list[str]

        :param host: Snapcast server (default: default configured host)
        :type host: str

        :param port: Snapcast server port (default: default configured port)
        :type port: int
        """

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            group = self._get_group(sock, group)
            request = {
                'id': self._get_req_id(),
                'jsonrpc':'2.0',
                'method': 'Group.SetClients',
                'params': {
                    'id': group['id'],
                    'clients': []
                }
            }

            for client in clients:
                client = self._get_client(sock, client)
                request['params']['clients'].append(client['id'])

            self._send(sock, request)
            return self._recv(sock)
        finally:
            try: sock.close()
            except: pass


    @action
    def group_set_stream(self, group, stream_id, host=None, port=None):
        """
        Sets the active stream for a group.

        :param group: Group name or ID
        :type group: str

        :param stream_id: Stream ID
        :type stream_id: str

        :param host: Snapcast server (default: default configured host)
        :type host: str

        :param port: Snapcast server port (default: default configured port)
        :type port: int
        """

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            group = self._get_group(sock, group)
            request = {
                'id': self._get_req_id(),
                'jsonrpc':'2.0',
                'method': 'Group.SetStream',
                'params': {
                    'id': group['id'],
                    'stream_id': stream_id,
                }
            }

            self._send(sock, request)
            return self._recv(sock)
        finally:
            try: sock.close()
            except: pass


    @action
    def get_backend_hosts(self):
        """
        :return: A dict with the Snapcast hosts configured on the backend
            in the format host -> port
        """

        hosts = {}
        for i in range(len(self.backend_hosts)):
            hosts[self.backend_hosts[i]] = self.backend_ports[i]
        return hosts

    @action
    def get_playing_streams(self, exclude_local=False):
        """
        Returns the remote streams configured in the `music.snapcast` backend
        that are currently active and unmuted.

        :param exclude_local: Exclude localhost connections (default: False)
        :type exclude_local: bool

        :returns: dict with the host->port mapping.

        Example::

            {
                "hosts": {
                    "server_1": 1705,
                    "server_2": 1705,
                    "server_3": 1705
                }
            }

        """

        backend_hosts = self.get_backend_hosts().output
        playing_hosts = {}

        def _worker(host, port):
            try:
                if exclude_local and (host == 'localhost'
                                    or  host == Config.get('device_id')):
                    return

                server_status = self.status(host=host, port=port).output
                client_status = self.status(host=host, port=port,
                                            client=Config.get('device_id')).output

                if client_status.get('config', {}).get('volume', {}).get('muted'):
                    return

                group = [g for g in server_status.get('groups', {})
                        if g.get('id') == client_status.get('group_id')].pop(0)

                if group.get('muted'):
                    return

                stream = [s for s in server_status.get('streams')
                        if s.get('id') == group.get('stream_id')].pop(0)

                if stream.get('status') != 'playing':
                    return

                playing_hosts[host] = port
            except Exception as e:
                self.logger.warning(('Error while retrieving the status of ' +
                                     'Snapcast host at {}:{}: {}').format(
                                         host, port, str(e)))

        workers = []

        for host, port in backend_hosts.items():
            w = threading.Thread(target=_worker, args=(host,port))
            w.start()
            workers.append(w)

        while workers:
            w = workers.pop()
            w.join()

        return {'hosts': playing_hosts}


# vim:sw=4:ts=4:et:



import datetime
from typing import List

from linode_api4.objects.linode import Instance, Config, Disk, Backup, Image, Kernel, Type

from platypush.message import Mapping
from platypush.message.response import Response


class LinodeResponse(Response):
    pass


class LinodeConfigModel(Mapping):
    def __init__(self, config: Config):
        super().__init__()
        self.comments = config.comments
        self.created = config.created
        self.helpers = config.helpers.dict
        self.id = config.id
        self.initrd = config.initrd
        self.kernel = dict(LinodeKernelModel(config.kernel))
        self.label = config.label
        self.linode_id = config.linode_id
        self.memory_limit = config.memory_limit
        self.parent_id_name = config.parent_id_name
        self.root_device = config.root_device
        self.run_level = config.run_level
        self.updated = datetime.datetime.fromisoformat(config.updated)
        self.virt_mode = config.virt_mode


class LinodeKernelModel(Mapping):
    def __init__(self, kernel: Kernel):
        super().__init__()
        self.architecture = kernel.architecture
        self.created = kernel.created
        self.deprecated = kernel.deprecated
        self.description = kernel.description
        self.id = kernel.id
        self.kvm = kernel.kvm
        self.label = kernel.label
        self.version = kernel.version
        self.xen = kernel.xen


class LinodeBackupModel(Mapping):
    def __init__(self, backup: Backup):
        super().__init__()
        self.created = backup.created
        self.disks = {
            disk.label: {
                'label': disk.label,
                'size': disk.size,
                'filesystem': disk.filesystem,
            }
            for disk in backup.disks
        }

        self.duration = backup.duration
        self.finished = backup.finished
        self.id = backup.id
        self.label = backup.label
        self.linode_id = backup.linode_id
        self.message = backup.message
        self.parent_id_name = backup.parent_id_name
        self.country = backup.region.country
        self.status = backup.status
        self.type = backup.type
        self.updated = backup.updated


class LinodeDiskModel(Mapping):
    def __init__(self, disk: Disk):
        super().__init__()
        self.created = disk.created
        self.filesystem = disk.filesystem
        self.id = disk.id
        self.label = disk.label
        self.linode_id = disk.linode_id
        self.parent_id_name = disk.parent_id_name
        self.size = disk.size
        self.status = disk.status
        self.updated = disk.updated


class LinodeImageModel(Mapping):
    def __init__(self, image: Image):
        super().__init__()
        self.created = image.created
        self.created_by = image.created_by
        self.deprecated = image.deprecated
        self.description = image.description
        self.is_public = image.is_public
        self.label = image.label
        self.size = image.size
        self.status = image.status
        self.type = image.type
        self.vendor = image.vendor


class LinodeTypeModel(Mapping):
    # noinspection PyShadowingBuiltins
    def __init__(self, type: Type):
        super().__init__()
        self.disk = type.disk
        self.id = type.id
        self.label = type.label
        self.memory = type.memory
        self.network_out = type.network_out
        self.price = type.price.dict
        self.transfer = type.transfer
        self.type_class = type.type_class
        self.vcpus = type.vcpus


class LinodeInstanceModel(Mapping):
    def __init__(self, node: Instance):
        super().__init__()
        self.label = node.label
        self.status = node.status
        self.alerts = node.alerts.dict
        self.available_backups = [
            dict(LinodeBackupModel(backup))
            for backup in node.available_backups.automatic
        ],

        self.backups = {
            'enabled': node.backups.enabled,
            'schedule': node.backups.schedule.dict,
            'last_successful': datetime.datetime.fromisoformat(node.backups.last_successful),
        }

        self.configs = {config.label: dict(LinodeConfigModel(config)) for config in node.configs}
        self.disks = {disk.label: dict(LinodeDiskModel(disk)) for disk in node.disks}
        self.group = node.group
        self.hypervisor = node.hypervisor
        self.id = node.id
        self.image = LinodeImageModel(node.image)
        self.country = node.region.country
        self.specs = node.specs.dict
        self.tags = node.tags
        self.transfer = node.transfer.dict
        self.type = dict(LinodeTypeModel(node.type))
        self.updated = node.updated


class LinodeInstanceResponse(LinodeResponse):
    def __init__(self, instance: Instance, *args, **kwargs):
        super().__init__(*args, output={
            'instance': dict(LinodeInstanceModel(instance))
        }, **kwargs)


class LinodeInstancesResponse(LinodeResponse):
    def __init__(self,
                 instances: List[Instance],
                 *args, **kwargs):
        super().__init__(*args, output={
            'instances': {instance.label: dict(LinodeInstanceModel(instance)) for instance in instances},
        }, **kwargs)


# vim:sw=4:ts=4:et:



import os

from platypush.plugins import Plugin, action


class MlModel:
    def __init__(self, model_file, classes=None):
        import cv2

        self.model_file = os.path.abspath(os.path.expanduser(model_file))
        self.classes = classes or []
        self.model = cv2.dnn.readNet(model_file)

    def predict(self, img, resize=None, color_convert=None):
        import cv2
        import numpy as np

        if isinstance(img, str):
            img = cv2.imread(os.path.abspath(os.path.expanduser(img)))

        if color_convert:
            if isinstance(color_convert, str):
                color_convert = getattr(cv2, color_convert)

            img = cv2.cvtColor(img, color_convert)

        if resize:
            img = cv2.dnn.blobFromImage(img, size=tuple(resize), mean=0.5)
        else:
            img = cv2.dnn.blobFromImage(img, mean=0.5)

        self.model.setInput(img)
        output = self.model.forward()
        prediction = int(np.argmax(output))

        if self.classes:
            prediction = self.classes[prediction]

        return prediction


class MlCvPlugin(Plugin):
    """
    Plugin to train and make computer vision predictions using machine learning models.

    Requires:

        * **numpy** (``pip install numpy``)
        * **opencv** (``pip install cv2``)

    Also make sure that your OpenCV installation comes with the ``dnn`` module. To test it::

        >>> import cv2.dnn

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.models = {}

    @action
    def predict(self, img, model_file, classes=None, resize=None, color_convert=None):
        """
        Make predictions for an input image using a model file. Supported model formats include all the
        types supported by cv2.dnn (currently supported: Caffe, TensorFlow, Torch, Darknet, DLDT).

        :param model_file: Path to the model file
        :param img: Path to the image
        :param classes: List of string labels associated with the output values (e.g. ['negative', 'positive']).
            If not set then the index of the output neuron with highest value will be returned.
        :param resize: Tuple or list with the resize factor to be applied to the image before being fed to
            the model (default: None)
        :param color_convert: Color conversion to be applied to the image before being fed to the model.
            It points to a cv2 color conversion constant (e.g. ``cv2.COLOR_BGR2GRAY``) and it can be either
            the constant value itself or a string (e.g. 'COLOR_BGR2GRAY').
        """

        model_file = os.path.abspath(os.path.expanduser(model_file))

        if model_file not in self.models:
            self.models[model_file] = MlModel(model_file, classes=classes)

        return self.models[model_file].predict(img, resize=resize, color_convert=color_convert)


# vim:sw=4:ts=4:et:
